// Mitsuba includes
#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/logger.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/hw/basicshader.h>

// STL includes
#include <fstream>
#include <sstream>

/* Helping functions
 */
template<typename T>
inline T _clamp(T x, T m, T M) {
   return std::min<T>(std::max<T>(x, m), M);
}
template<typename T>
inline T sqr(T x) {
   return x*x;
}

/* On OSX, I cannot compile Mitsuba with full c++11. Some features are thus
 * missing. Such as `to_string` ...
 */
#if defined(__APPLE__)
namespace std {
template <typename T>
std::string to_string(const T& value) {
    std::stringstream ss;
    ss << value;
    return ss.str();
}
}
#endif

/* Roughness to linear space conversions
 */
//#define USE_BEST_FIT
inline float roughnessToVariance(float a) {
#ifdef USE_BEST_FIT
   a = _clamp<float>(a, 0.0, 0.9999);
   float a3 = powf(a, 1.1);
   return a3 / (1.0f - a3);
#else
   return a / (1.0f-a);
#endif
}
inline float varianceToRoughness(float v) {
#ifdef USE_BEST_FIT
   return powf(v / (1.0f + v), 1.0f/1.1f);
#else
   return v / (1.0f+v);
#endif
}
inline float gToVariance(float g) {
    assert(g > 0.0f && g <= 1.0f);
    return pow((1.0f-g)/g, 0.8f) / (1.0f+g);
}

MTS_NAMESPACE_BEGIN

void parseLayers(const Properties &props,
                 int& nb_layers,
                 std::vector<Spectrum>& m_etas,
                 std::vector<Spectrum>& m_kappas,
                 std::vector<Float>&    m_alphas,
                 std::vector<Float>&    m_depths,
                 std::vector<Spectrum>& m_sigma_s,
                 std::vector<Spectrum>& m_sigma_a,
                 std::vector<Float>&    m_gs) {

    /* Parse a layered structure */
    nb_layers = props.getInteger("nb_layers", 0);
    //AssertEx(nb_layers > 0, "layered must have at least one layer");

    // Add the external IOR
    Float extEta = lookupIOR(props, "extEta", "air");
    m_etas.push_back(Spectrum(extEta));
    m_kappas.push_back(Spectrum(0.0));

    // Add the layers IOR and interfaces
    for(int k=0; k<nb_layers; ++k) {
        std::string index = std::to_string(k);
        std::string name;

        // Adding a rough surface parameters
        name = std::string("eta_") + index;
        Spectrum eta_k = props.getSpectrum(name, Spectrum(1.0f));

        name = std::string("kappa_") + index;
        Spectrum kappa_k = props.getSpectrum(name, Spectrum(0.0f));

        name = std::string("alpha_") + index;
        Float alpha_k = props.getFloat(name, 0.0f);

        // Adding the participating media parameters. This should overloads the
        // already defined parameters in computation. This has to be tested
        // using the depth > 0.0 test.
        name = std::string("depth_") + index;
        Float depth = props.getFloat(name, 0.0f);

        name = std::string("sigmas_") + index;
        Spectrum sigma_s = props.getSpectrum(name, Spectrum(0.0f));

        name = std::string("sigmaa_") + index;
        Spectrum sigma_a = props.getSpectrum(name, Spectrum(0.0f));

        name = std::string("g_") + index;
        Float g = props.getFloat(name, 0.9f);

        // Propagate IORs when a volume is set
        // And compute the scattering parameter.
        if(depth > 0.0f) {
            eta_k   = m_etas.back();
            kappa_k = m_kappas.back();
            alpha_k = gToVariance(g);
        }

        // Update roughness
        m_etas.push_back(eta_k);
        m_kappas.push_back(kappa_k);
        m_alphas.push_back(alpha_k);

        // Update the media
        m_depths.push_back(depth);
        m_sigma_s.push_back(sigma_s);
        m_sigma_a.push_back(sigma_a);
        m_gs.push_back(g);

    }

    // Print the resulting microstructure
    SLog(EInfo, "");
    SLog(EInfo, "Adding Exterior IOR");
    SLog(EInfo, " + n = %s", m_etas[0].toString().c_str());
    SLog(EInfo, " + k = %s", m_kappas[0].toString().c_str());
    SLog(EInfo, "");

    for(int k=0; k<nb_layers; ++k) {
        SLog(EInfo, "Adding layer %d", k);
        if(m_depths[k] > 0.0f) {
           SLog(EInfo, " + d  = %f", m_depths[k]);
           SLog(EInfo, " + ss = %s", m_sigma_s[k].toString().c_str());
           SLog(EInfo, " + sa = %s", m_sigma_a[k].toString().c_str());
           SLog(EInfo, " + g  = %f", m_gs[k]);
           SLog(EInfo, "");
        } else {
           SLog(EInfo, " + n = %s", m_etas[k+1].toString().c_str());
           SLog(EInfo, " + k = %s", m_kappas[k+1].toString().c_str());
           SLog(EInfo, " + a = %f", m_alphas[k]);
           SLog(EInfo, "");
        }
    }
}

void parseLayers(const Properties &props,
                 int& nb_layers,
                 std::vector<ref<const Texture> >& m_tex_etas,
                 std::vector<ref<const Texture> >& m_tex_kappas,
                 std::vector<ref<const Texture> >& m_tex_alphas,
                 std::vector<Float>&    m_depths,
                 std::vector<Spectrum>& m_sigma_s,
                 std::vector<Spectrum>& m_sigma_a,
                 std::vector<Float>&    m_gs) {

    /* Parse a layered structure */
    nb_layers = props.getInteger("nb_layers", 0);
    //AssertEx(nb_layers > 0, "layered must have at least one layer");

    // Add the external IOR
    Float extEta = lookupIOR(props, "extEta", "air");
    m_tex_etas.push_back(new ConstantSpectrumTexture(Spectrum(extEta)));
    m_tex_kappas.push_back(new ConstantSpectrumTexture(Spectrum(0.0)));

    // Add the layers IOR and interfaces
    for(int k=0; k<nb_layers; ++k) {
        std::string index = std::to_string(k);
        std::string name;

        // Adding a rough surface parameters
        name = std::string("eta_") + index;
        Spectrum eta_k = props.getSpectrum(name, Spectrum(1.0f));

        name = std::string("kappa_") + index;
        Spectrum kappa_k = props.getSpectrum(name, Spectrum(0.0f));

        name = std::string("alpha_") + index;
        Float alpha_k = props.getFloat(name, 0.0f);

        // Adding the participating media parameters. This should overloads the
        // already defined parameters in computation. This has to be tested
        // using the depth > 0.0 test.
        name = std::string("depth_") + index;
        Float depth = props.getFloat(name, 0.0f);

        name = std::string("sigmas_") + index;
        Spectrum sigma_s = props.getSpectrum(name, Spectrum(0.0f));

        name = std::string("sigmaa_") + index;
        Spectrum sigma_a = props.getSpectrum(name, Spectrum(0.0f));

        name = std::string("g_") + index;
        Float g = props.getFloat(name, 0.9f);

        // Propagate IORs when a volume is set
        // And compute the scattering parameter.
        if(depth > 0.0f) {
            eta_k   = m_tex_etas.back()->getAverage();
            kappa_k = m_tex_kappas.back()->getAverage();
            alpha_k = gToVariance(g);
        }

        // Update roughness
        m_tex_etas.push_back(new ConstantSpectrumTexture(eta_k));
        m_tex_kappas.push_back(new ConstantSpectrumTexture(kappa_k));
        m_tex_alphas.push_back(new ConstantFloatTexture(alpha_k));

        // Update the media
        m_depths.push_back(depth);
        m_sigma_s.push_back(sigma_s);
        m_sigma_a.push_back(sigma_a);
        m_gs.push_back(g);

    }

    // Print the resulting microstructure
    SLog(EInfo, "");
    SLog(EInfo, "Adding Exterior IOR");
    SLog(EInfo, " + n = %s", m_tex_etas[0]->getAverage().toString().c_str());
    SLog(EInfo, " + k = %s", m_tex_kappas[0]->getAverage().toString().c_str());
    SLog(EInfo, "");

    for(int k=0; k<nb_layers; ++k) {
        SLog(EInfo, "Adding layer %d", k);
        if(m_depths[k] > 0.0f) {
            SLog(EInfo, " + d  = %f", m_depths[k]);
            SLog(EInfo, " + ss = %s", m_sigma_s[k].toString().c_str());
            SLog(EInfo, " + sa = %s", m_sigma_a[k].toString().c_str());
            SLog(EInfo, " + g  = %f", m_gs[k]);
            SLog(EInfo, "");
        } else {
            SLog(EInfo, " + n = %s", m_tex_etas[k+1]->getAverage().toString().c_str());
            SLog(EInfo, " + k = %s", m_tex_kappas[k+1]->getAverage().toString().c_str());
            SLog(EInfo, " + a = %f", m_tex_alphas[k]->getAverage().average());
            SLog(EInfo, "");
        }
    }
}


/* Sample the HG phase function
 */
Vector3 HG_Sample(const Vector3& wi, Float g, const Vector2& uv) {

   Float cosTheta;
   if (std::abs(g) < Epsilon) {
      cosTheta = 1 - 2*uv.x;
   } else {
      Float sqrTerm = (1 - g * g) / (1 - g + 2 * g * uv.x);
      cosTheta = (1 + g * g - sqrTerm * sqrTerm) / (2 * g);
   }

   Float sinTheta = math::safe_sqrt(1.0f-cosTheta*cosTheta),
         sinPhi, cosPhi;

   math::sincos(2*M_PI*uv.y, &sinPhi, &cosPhi);

   const Vector ws(sinTheta * cosPhi, sinTheta * sinPhi, cosTheta);
   return Frame(wi).toWorld(ws);
}


#define LINEAR_INTERPOLATION

/* FGD structure to linearly interpolate the 4D datafile.
 */
struct FGD {
    FGD() {}
    FGD(const fs::path &path) {
        fs::ifstream in(path);
        // Warn if not loaded
        if(in.bad() || in.fail()) {
            SLog(EError, "Unable to load file: %s", path.string().c_str());
        } else {
            SLog(EInfo, "Loading FGD file: %s", path.string().c_str());
        }
        this->load(in);
    }
    void load(fs::ifstream& in) {
        // Read sizes
        in.read((char*)&Nt, sizeof(int));
        in.read((char*)&Na, sizeof(int));
        in.read((char*)&Nn, sizeof(int));
        in.read((char*)&Nk, sizeof(int));
        SLog(EInfo, "Loading FGD texture of dimension %dx%dx%dx%d", Nt, Na, Nn, Nk);
        sizes[0] = Nt;
        sizes[1] = Na;
        sizes[2] = Nn;
        sizes[3] = Nk;

        // Read data range (min / max)
        float mM[8];
        in.read((char*)mM, 8*sizeof(float));
        tm = mM[0]; tM = mM[1];
        am = mM[2]; aM = mM[3];
        nm = mM[4]; nM = mM[5];
        km = mM[6]; kM = mM[7];
        SLog(EInfo, "Range [%f, %f; %f, %f; %f, %f; %f, %f]", tm, tM, am, aM, nm, nM, km, kM);

        // Read data
        const int size = Nt*Na*Nn*Nk;
        buff.assign(size, 0.0f);
        in.read((char*)buff.data(), size*sizeof(float));

    }

    inline Float operator() (Float t, Float a, Float n, Float k) const {

       // Floating point index
       Float ta = Nt * (t - tm) / (tM -tm);
       Float aa = Na * (a - am) / (aM -am);
       Float na = Nn * (n - nm) / (nM -nm);
       Float ka = Nk * (k - km) / (kM -km);

       // Integer index
       int ti = floor(ta);
       int ai = floor(aa);
       int ni = floor(na);
       int ki = floor(ka);

       // Ensure the indexes stays in the limits
       ti = _clamp<int>(ti, 0, Nt-1);
       ai = _clamp<int>(ai, 0, Na-1);
       ni = _clamp<int>(ni, 0, Nn-1);
       ki = _clamp<int>(ki, 0, Nk-1);

       //*
       // Clamp the interpolation weights
       Float alphas[4] = {ta-ti, aa-ai, na-ni, ka-ki};
       for(int i=0; i<4; ++i) {
          alphas[i] = _clamp<Float>(alphas[i], 0.0f, 1.0f);
       }

       // Index of the middle point
       const int indices[4] = {ti, ai, ni, ki};
       const int index = ki + Nk*(ni + Nn*(ai + Na*ti));

#ifdef LINEAR_INTERPOLATION
       // Result vector
       Float v = 0.0f;

       // For every possible combinaison of index shift per dimension,
       // fetch the value in memory and do linear interpolation.
       // We fetch using shift of 0 and 1.
       //
       //     v(i+di, j+di, k+dk, l+dl),  where dk in [0,1]
       //
       const unsigned int D = pow(2, 4);
       for(unsigned int d=0; d<D; ++d) {

          Float alpha = 1.0; // Global alpha
          int   cid_s = 0;   // Id shift

          // Evaluate the weight of the sample d which correspond to
          // one for the shifted configuration:
          // The weight is the product of the weights per dimension.
          //
          for(int i=0; i<4; ++i) {
             bool  bitset = ((1 << i) & d);
             Float calpha =  (bitset) ? alphas[i] : 1.0-alphas[i];

             // Correct the shift to none if we go out of the grid
             if(indices[i] + 1 >= sizes[i]) {
                bitset = false;
             }

             alpha *= calpha;
             cid_s = cid_s*sizes[i] + ((bitset) ? 1 : 0);
          }

          v += alpha * buff[index+cid_s];
       }
       return v;
#else
       return buff[index];
#endif
    }
    Spectrum operator() (Float t, Float a, const Spectrum& n, const Spectrum& k) const {
       Spectrum s;
       for(int i=0; i<SPECTRUM_SAMPLES; ++i) {
          s[i] = (*this)(t, a, n[i], k[i]);
       }
       return s;
    }

    std::vector<float> buff;
    int sizes[4];
    int Nt, Na, Nn, Nk;
    float tm, tM, aM, am, nm, nM, km, kM;
};


struct TIR {
    TIR() {}
    TIR(const fs::path &path) {
        fs::ifstream in(path);
        // Warn if not loaded
        if(in.bad() || in.fail()) {
            SLog(EError, "Unable to load TIR file: %s", path.string().c_str());
        } else {
            SLog(EInfo, "Loading TIR file: %s", path.string().c_str());
        }
        this->load(in);
    }
    void load(fs::ifstream& in) {
        // Read sizes
        in.read((char*)&Nt, sizeof(int));
        in.read((char*)&Na, sizeof(int));
        in.read((char*)&Nn, sizeof(int));
        SLog(EInfo, "Loading TIR texture of dimension %dx%dx%dx", Nt, Na, Nn);
        sizes[0] = Nt;
        sizes[1] = Na;
        sizes[2] = Nn;

        // Read data range (min / max)
        float mM[6];
        in.read((char*)mM, 6*sizeof(float));
        tm = mM[0]; tM = mM[1];
        am = mM[2]; aM = mM[3];
        nm = mM[4]; nM = mM[5];
        SLog(EInfo, "Range [%f, %f; %f, %f; %f, %f]", tm, tM, am, aM, nm, nM);

        // Read data
        const int size = Nt*Na*Nn;
        buff.assign(size, 0.0f);
        in.read((char*)buff.data(), size*sizeof(float));

    }

    inline float operator() (float t, float a, float n) const {

       // Floating point index
       float ta = Nt * (t - tm) / (tM -tm);
       float aa = Na * (a - am) / (aM -am);
       float na = Nn * (n - nm) / (nM -nm);

       // Integer index
       int ti = floor(ta);
       int ai = floor(aa);
       int ni = floor(na);

       // Ensure the indexes stays in the limits
       ti = _clamp<int>(ti, 0, Nt-1);
       ai = _clamp<int>(ai, 0, Na-1);
       ni = _clamp<int>(ni, 0, Nn-1);

       //*
       // Clamp the interpolation weights
       float alphas[3] = {ta-ti, aa-ai, na-ni};
       for(int i=0; i<3; ++i) {
          alphas[i] = _clamp<float>(alphas[i], 0.0f, 1.0f);
       }

       // Index of the middle point
       const int indices[3] = {ti, ai, ni};
       const int index = ni + Nn*(ai + Na*ti);


#ifdef LINEAR_INTERPOLATION
       // Result vector and norm
       float v = 0.0f;
       float V = 0.0f;

       // For every possible combinaison of index shift per dimension,
       // fetch the value in memory and do linear interpolation.
       // We fetch using shift of 0 and 1.
       //
       //     v(i+di, j+di, k+dk, l+dl),  where dk in [0,1]
       //
       const unsigned int D = pow(2, 3);
       for(unsigned int d=0; d<D; ++d) {

          float alpha = 1.0; // Global alpha
          int   cid_s = 0;   // Id shift

          // Evaluate the weight of the sample d which correspond to
          // one for the shifted configuration:
          // The weight is the product of the weights per dimension.
          //
          for(int i=0; i<3; ++i) {
             bool  bitset = ((1 << i) & d);
             float calpha =  (bitset) ? alphas[i] : 1.0-alphas[i];

             // Correct the shift to none if we go out of the grid
             if(indices[i] + 1 >= sizes[i]) {
                bitset = false;
             }

             alpha *= calpha;
             cid_s = cid_s*sizes[i] + ((bitset) ? 1 : 0);
          }

          const float tmp = buff[index+cid_s];
          if(! std::isnan(tmp)) {
            v += alpha * tmp;
            V += alpha;
          }
       }

       return v;
#else
       const int index = ki + Nk*(ni + Nn*(ai + Na*ti));
       return buff[index];
#endif
    }

    std::vector<float> buff;
    int sizes[3];
    int Nt, Na, Nn;
    float tm, tM, aM, am, nm, nM;
};

MTS_NAMESPACE_END
