// Mitsuba includes
#include <mitsuba/core/fresolver.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/hw/basicshader.h>

// Local includes
#include "microfacet.h"
#include "ior.h"

// Local includes
#include "layered.hpp"

MTS_NAMESPACE_BEGIN

class LayeredDielectric : public BSDF {
public:
    LayeredDielectric(const Properties &props) : BSDF(props) {
        ref<FileResolver> fResolver = Thread::getThread()->getFileResolver();

        /* Parse a layered structure */
        parseLayers(props, nb_layers, m_tex_etas, m_tex_kappas, m_tex_alphas,
                                      m_depths, m_sigma_s, m_sigma_a, m_gs);

        // Load the FGD and TIR textures
        m_useFGD = props.getBoolean("useFGD", false);
        if(m_useFGD) {
            m_FGD = FGD(fResolver->resolve("data/FGD.bin").string());
        }
        m_useTIR = props.getBoolean("useTIR", false);
        if(m_useTIR){
            m_TIR = TIR(fResolver->resolve("data/TIR.bin").string());
        }

        // Ensure that no layer has non-zero kappa
        for(int i=0; i<nb_layers; ++i) {
            Assert(m_tex_kappas[i]->getAverage().isZero());
        }
    }

    LayeredDielectric(Stream *stream, InstanceManager *manager)
     : BSDF(stream, manager) {
        /*
        m_type = (MicrofacetDistribution::EType) stream->readUInt();
        m_sampleVisible = stream->readBool();
        m_alphaU = static_cast<Texture *>(manager->getInstance(stream));
        m_alphaV = static_cast<Texture *>(manager->getInstance(stream));
        m_specularReflectance = static_cast<Texture *>(manager->getInstance(stream));
        m_eta = Spectrum(stream);
        m_k = Spectrum(stream);
         */
        configure();
    }

    void serialize(Stream *stream, InstanceManager *manager) const {
        BSDF::serialize(stream, manager);

        /*
        stream->writeUInt((uint32_t) m_type);
        stream->writeBool(m_sampleVisible);
        manager->serialize(stream, m_alphaU.get());
        manager->serialize(stream, m_alphaV.get());
        manager->serialize(stream, m_specularReflectance.get());
        m_eta.serialize(stream);
        m_k.serialize(stream);
        */
    }

    void configure() {
        unsigned int extraFlags = 0;
        /*
        if (!m_alphaU->isConstant() || !m_alphaV->isConstant() ||
            !m_specularReflectance->isConstant())
            extraFlags |= ESpatiallyVarying;
         */

        m_components.clear();
		m_components.push_back(EGlossyReflection | EFrontSide
			| EBackSide | EUsesSampler | ESpatiallyVarying);
		m_components.push_back(EGlossyTransmission | EFrontSide
			| EBackSide | EUsesSampler | ENonSymmetric | ESpatiallyVarying);

        m_usesRayDifferentials = false;
        /*
            m_alphaU->usesRayDifferentials() ||
            m_alphaV->usesRayDifferentials() ||
            m_specularReflectance->usesRayDifferentials();
         */

        BSDF::configure();
    }

    /*
     */
    void evalFresnel(Float ct, Float alpha, Float eta,
                     Spectrum& Rij, Spectrum& Tij) const {

        if(m_useFGD) {
            Rij = m_FGD(ct, alpha, eta, 0.0f) * Spectrum(1.0f);
        } else {
            Rij = fresnelDielectricExt(ct, eta) * Spectrum(1.0f);
        }
        Tij = Spectrum(1.0) - Rij;
    }

    /* Trace moment analysis in the layer structure for a given input angle
     */
    void computeAddingDoubling(const Vector& wi,
                               const Spectrum* m_etas, const Spectrum* m_kappas, const Float* m_alphas,
                               Spectrum* coeffs, Vector3* dirs, Float* alphas) const {

        /* Constants */
        static const Vector3 N = Vector3(0, 0, 1);

        // Variables
        Spectrum R0i(0.0f), Ri0(0.0f), T0i(1.0f), Ti0(1.0f);
        Vector3 wt = -wi; Vector3 wr;
        float s_r0i = 0.0f, s_ri0=0.0f, s_t0i=0.0f, s_ti0=0.0f;
        float j0i=1.0f, ji0=1.0f;

        // Iterate over the layers
        for(int i=0; i<nb_layers; ++i) {

            /* Extract layer data */
            float eta_1   = m_etas[i].average();
            float eta_2   = m_etas[i+1].average();
            float eta     = eta_2 / eta_1;
            float ieta    = eta_1 / eta_2;
            float alpha   = m_alphas[i];
            float n12     = eta;
            float depth   = m_depths[i];

            Spectrum R12, T12, R21, T21;
            float s_r12=0.0f, s_r21=0.0f, s_t12=0.0f, s_t21=0.0f, j12=1.0f, j21=1.0f;
            if(depth > 0.0f) {
                /* Mean doesn't change with volumes */
                wr = wt;

                /* Evaluate transmittance */
                const Spectrum sigma_t   = m_sigma_a[i] + m_sigma_s[i];
                const float    opt_depth = depth/fabs(wt.z);
                T12 = (Spectrum(1.0f) + m_sigma_s[i]*opt_depth) * (- (opt_depth) * sigma_t).exp();
                T21 = T12;
                R12 = Spectrum(0.0f);
                R21 = Spectrum(0.0f);

                /* Fetch precomputed variance for HG phase function */
                s_t12 = alpha;
                s_t21 = alpha;

            } else {
                /* Evaluate the mean direction for refraction */
                wr = refract(-wt, N, n12);

                /* Ray is not block by conducting interface or total reflection */
                const bool has_transmissive = !wr.isZero();

                /* Evaluate interface variance term */
                s_r12 = roughnessToVariance(alpha);
                s_r21 = s_r12;

                /* For dielectric interfaces, evaluate the transmissive roughnesses */
                if(has_transmissive) {
                    const float ctt = fabs(wr.z);
                    const float cti = fabs(wt.z);
                    s_t12 = roughnessToVariance(alpha * 0.5f * fabs(ctt*n12 - cti)/(ctt*n12));
                    s_t21 = roughnessToVariance(alpha * 0.5f * fabs(cti/n12 - ctt)/(cti/n12));
                      j12 = (ctt/cti)*n12; // Scale due to the interface
                      j21 = 1.0f/j12;
                }

                /* Evaluate FGD using a modified roughness accounting for top layers */
                auto temp_alpha = varianceToRoughness(s_t0i + s_r12);

                /* Evaluate r12, r21, t12, t21 */
                if(has_transmissive) {
                    evalFresnel(wt.z, temp_alpha, 1.0/eta, R12, T12);
                    // R12 = 0.0*R12;
                    R21 = R12;
                    T12 = T12;
                    T21 = T12 * (n12*n12);
                    T12 = T12 / (n12*n12);
                } else {
                    R12 = Spectrum(0.0f);
                    R21 = Spectrum(0.0f);
                    T21 = Spectrum(0.0f);
                    T12 = Spectrum(0.0f);
                }

                /* Evaluate TIR using the decoupling approximation */
                if(i > 0 && m_useTIR) {
                    const float cti = fabs(wt.z);
                    Spectrum eta_0   = (i==0) ? m_etas[0] : m_etas[i-1];
                    float n10        = (eta_0/eta_1).average();

                    const float _TIR  = m_TIR(cti, temp_alpha, n10);
                    Ri0 += _TIR * Ti0;
		            for(int llid=0; llid<3; ++llid) { Ri0[llid] = fmin(Ri0[llid], 1.0f); }
                    Ti0 *= 1.0f - _TIR;
                }
            }

            /* Evaluate the adding operators on the energy */
            const Spectrum denom = (Spectrum(1.0f) - Ri0*R12);

            /* Multiple scattering forms */
            const Spectrum m_R0i = (denom.average() <= 0.0f)? Spectrum(0.0f) : (T0i*R12*Ti0) / denom;
            const Spectrum m_Ri0 = (denom.average() <= 0.0f)? Spectrum(0.0f) : (T21*Ri0*T12) / denom;
            const Spectrum m_Rr  = (denom.average() <= 0.0f)? Spectrum(0.0f) : (Ri0*R12) / denom;
            const float m_r0i    = m_R0i.average();
            const float m_ri0    = m_Ri0.average();
            const float m_rr     = m_Rr.average();
            const float r21      = R21.average();

            /* Evaluate the adding operator on the energy */
            const Spectrum e_R0i = R0i + m_R0i;
            const Spectrum e_T0i = (T0i*T12) / denom;
            const Spectrum e_Ri0 = R21 + (T21*Ri0*T12) / denom;
            const Spectrum e_Ti0 = (T21*Ti0) / denom;

            /* Scalar form for the energy */
            const float r0i   = R0i.average();
            const float e_r0i = e_R0i.average();
            const float e_ri0 = e_Ri0.average();

            /* Evaluate the adding operator on the normalized variance */
            float _s_r0i = (r0i*s_r0i + m_r0i*(s_ti0 + j0i*(s_t0i + s_r12 + m_rr*(s_r12+s_ri0))));
            float _s_t0i = j12*s_t0i + s_t12 + j12*(s_r12 + s_ri0)*m_rr;
            float _s_ri0 = (r21*s_r21 + m_ri0*(s_t12 + j12*(s_t21 + s_ri0 + m_rr*(s_r12+s_ri0))));
            float _s_ti0 = ji0*s_t21 + s_ti0 + ji0*(s_r12 + s_ri0)*m_rr;
            _s_r0i = (e_r0i > 0.0f) ? _s_r0i/e_r0i : 0.0f;
            _s_ri0 = (e_ri0 > 0.0f) ? _s_ri0/e_ri0 : 0.0f;

            /* Update energy */
            R0i = e_R0i;
            T0i = e_T0i;
            Ri0 = e_Ri0;
            Ti0 = e_Ti0;

            /* Update mean */
            wt = wr;

            /* Update variance */
            s_r0i = _s_r0i;
            s_t0i = _s_t0i;
            s_ri0 = _s_ri0;
            s_ti0 = _s_ti0;

            /* Update jacobian */
            j0i *= j12;
            ji0 *= j21;

            /* Early stop */
            if(T0i.isZero()) {
                break;
            }
        }

        /* Store the coefficient and variance */
        coeffs[0] = R0i;
        dirs[0]   = wi;
        alphas[0] = varianceToRoughness(s_r0i);
        coeffs[1] = T0i;
        dirs[1]   = reflect(wt, N);
        alphas[1] = varianceToRoughness(s_t0i);
    }

    /* Evaluate the maps */
    inline void evalMaps(const BSDFSamplingRecord &bRec, bool upward,
                         Spectrum* m_etas, Spectrum* m_kappas,
                         Float* m_alphas) const {


        // Fetch air values
        upward = true;
        const int j = upward ? 0 : nb_layers;
        m_etas[j]   = m_tex_etas[0]->eval(bRec.its);
        m_kappas[j] = m_tex_kappas[0]->eval(bRec.its);

        // Fetch textured values
        for(int i=0; i<nb_layers; ++i) {
            const int j    = upward ? i : (nb_layers-1) - i;
            const int dt   = upward ? 1 : 0;
            m_etas[j+dt]   = m_tex_etas[i+1]->eval(bRec.its);
            m_kappas[j+dt] = m_tex_kappas[i+1]->eval(bRec.its);
            m_alphas[j]    = m_tex_alphas[i]->eval(bRec.its).average();
            // TODO Add other maps
        }
    }

    Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
        /* Stop if this component was not requested */
		if (measure != ESolidAngle || Frame::cosTheta(bRec.wi) == 0)
			return Spectrum(0.0f);

        /* Check consitency in flags */
        if (Frame::cosTheta(bRec.wi) * Frame::cosTheta(bRec.wo) > 0) {
			/* Stop if this component was not requested */
			if ((bRec.component != -1 && bRec.component != 0)
				|| !(bRec.typeMask & EGlossyReflection))
				return Spectrum(0.0f);

		} else {
			/* Stop if this component was not requested */
			if ((bRec.component != -1 && bRec.component != 1)
				|| !(bRec.typeMask & EGlossyTransmission))
				return Spectrum(0.0f);
		}

        /* If the ray travelling forward the structure */
        const bool upward = bRec.wi.z > 0.0f;

        /* Evaluate the maps */
        Float m_alphas[nb_layers];
        Spectrum m_kappas[nb_layers+1], m_etas[nb_layers+1];
        evalMaps(bRec, upward, m_etas, m_kappas, m_alphas);

        /* Evaluate the adding method to get coeffs and variances */
        Spectrum coeffs[2];
        Vector3  dirs[2];
        float    alphas[2];
        int nb_valid = 2;
        computeAddingDoubling(bRec.wi,
                              m_etas, m_kappas, m_alphas,
                              coeffs, dirs, alphas);

        Spectrum result(0.0f);
        for(int i=0; i<nb_valid; ++i) {
            /* Skip non possible configurations */
            if(dirs[i].z * bRec.wo.z < 0.0f) {
                continue;
            }

            /* Evaluate the half-vector and generate the NDF */
            const Vector3 H = normalize(dirs[i] + bRec.wo);
            MicrofacetDistribution distr(MicrofacetDistribution::EGGX,
                                        alphas[i], alphas[i], true);

            /* Evaluate the pdf */
            float DG = distr.eval(H) * distr.smithG1(math::signum(bRec.wo.z) * bRec.wo, H);
            result += coeffs[i] * (DG / (4.0f * fabs(dirs[i].z)));
        }

        return result;
    }

    Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {

        /* Check the measure */
 		if (measure != ESolidAngle)
			return 0.0f;

		/* Determine the type of interaction */
		bool hasReflection   = ((bRec.component == -1 || bRec.component == 0)
							  && (bRec.typeMask & EGlossyReflection)),
		     hasTransmission = ((bRec.component == -1 || bRec.component == 1)
							  && (bRec.typeMask & EGlossyTransmission)),
		     reflect         = Frame::cosTheta(bRec.wi)
				             * Frame::cosTheta(bRec.wo) > 0,
             upward          = Frame::cosTheta(bRec.wi) > 0;

        /* Check consitency in flags */
        if (reflect) {
			/* Zero probability if this component was not requested */
			if ((bRec.component != -1 && bRec.component != 0)
				|| !(bRec.typeMask & EGlossyReflection))
				return 0.0f;

		} else {
			/* Zero probability if this component was not requested */
			if ((bRec.component != -1 && bRec.component != 1)
				|| !(bRec.typeMask & EGlossyTransmission))
				return 0.0f;
		}

        /* Evaluate the maps */
        Float m_alphas[nb_layers];
        Spectrum m_kappas[nb_layers+1], m_etas[nb_layers+1];
        evalMaps(bRec, upward, m_etas, m_kappas, m_alphas);

        /* Evaluate the adding method to get coeffs and variances */
        Spectrum coeffs[2];
        Vector3  dirs[2];
        float    alphas[2];
        int nb_valid = 2;
        computeAddingDoubling(bRec.wi,
                              m_etas, m_kappas, m_alphas,
                              coeffs, dirs, alphas);

        /* Convert Spectral coefficients to float for pdf weighting */
        float pdf   = 0.0;
        float cum_w = 0.0;
        for(int i=0; i<nb_valid; ++i) {

            /* Evaluate weight */
            auto weight = coeffs[i].average();
            cum_w += weight;

            // Skip zero contributions
            if(coeffs[i].isZero() || dirs[i].z*bRec.wo.z < 0.0) {
                continue;
            }

            /* Evaluate Half vector & Generate NDF */
            Vector3 H = normalize(dirs[i] + bRec.wo);
            MicrofacetDistribution distr(MicrofacetDistribution::EGGX,
                                        alphas[i], alphas[i], true);

            /* Evaluate the pdf */
            float DG = distr.eval(H) * distr.smithG1(math::signum(dirs[i].z) * dirs[i], H) / (4.0f * fabs(Frame::cosTheta(dirs[i])));
            pdf += weight * DG;
            Assert(!std::isnan(pdf));
        }

         if(cum_w > 0.0f) {
            return pdf / cum_w;
         } else {
            return 0.0f;
         }
    }

    Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
        Float pdf;
        return this->sample(bRec, pdf, sample);
    }

    Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf, const Point2 &sample) const {

		/* Determine the type of interaction */
		bool hasReflection   = ((bRec.component == -1 || bRec.component == 0)
							  && (bRec.typeMask & EGlossyReflection)),
		     hasTransmission = ((bRec.component == -1 || bRec.component == 1)
							  && (bRec.typeMask & EGlossyTransmission)),
             upward          = Frame::cosTheta(bRec.wi) > 0;

		if (!hasReflection && !hasTransmission)
			return Spectrum(0.0f);

        /* Constant */
        const Vector3 N(0,0,1);

        /* Evaluate the maps */
        Float m_alphas[nb_layers];
        Spectrum m_kappas[nb_layers+1], m_etas[nb_layers+1];
        evalMaps(bRec, upward, m_etas, m_kappas, m_alphas);

        /* Evaluate the adding method to get coeffs and variances */
        Spectrum coeffs[2];
        Vector3  dirs[2];
        float    alphas[2];
        int nb_valid = 2;
        computeAddingDoubling(bRec.wi,
                              m_etas, m_kappas, m_alphas,
                              coeffs, dirs, alphas);

        /* Convert Spectral coefficients to floats to select BRDF lobe to sample */
        float weights[nb_valid];
        float cum_w = 0.0;
        for(int i=0; i<nb_valid; ++i) {
            weights[i] = coeffs[i].average();
            cum_w += weights[i];
        }
        if(cum_w <= 0.0) { return Spectrum(0.0f); }

        /* Select a random BRDF lobe */
        float sel_w = bRec.sampler->next1D() * cum_w - weights[0];
        int   sel_i = 0;
        for(sel_i=0; sel_w>0.0 && sel_i<nb_valid; sel_i++) {
            sel_w -= weights[sel_i+1];
        }

        /* Sample a microfacet normal */
        const MicrofacetDistribution distr(MicrofacetDistribution::EGGX,
                                           alphas[sel_i], alphas[sel_i], true);
        const Vector3 wi = dirs[sel_i];
        const Normal  m  = distr.sample(math::signum(Frame::cosTheta(wi)) * wi, sample, pdf);

        /* Check the direction of light */
        const bool is_reflect = wi.z > 0.0;

        /* Perfect specular reflection based on the microfacet normal */
        bRec.wo = reflect(wi, m);
        if((is_reflect && bRec.wi.z*bRec.wo.z <= 0.0f)  ||
           (!is_reflect && bRec.wi.z*bRec.wo.z >= 0.0f) || pdf <= 0.0f) {
            return Spectrum(0.0f);
        }

        bRec.eta = 1.0f;
        bRec.sampledComponent = (is_reflect) ? 0 : 1;
        bRec.sampledType = (is_reflect) ? EGlossyReflection : EGlossyTransmission;

        const Spectrum R = coeffs[sel_i] * distr.smithG1(bRec.wo, m);
        const float pdfW = weights[sel_i]/cum_w;
        pdf *= pdfW;
        return R/pdfW;

    }

    void addChild(const std::string &name, ConfigurableObject *child) {
        if (child->getClass()->derivesFrom(MTS_CLASS(Texture))) {
            std::string substr = name.substr(0, 3);
            if (substr == "eta") {
                int eta_index = atoi(name.substr(4).c_str());
                m_tex_etas[eta_index] = static_cast<Texture *>(child);
            } else if (substr == "kap") {
                int kappa_index = atoi(name.substr(6).c_str());
                m_tex_kappas[kappa_index] = static_cast<Texture *>(child);
            }  else if (substr == "alp") {
                int alpha_index = atoi(name.substr(6).c_str());
                m_tex_alphas[alpha_index] = static_cast<Texture *>(child);
            } else {
                // BSDF::addChild(name, child);
            }
        } else {
            BSDF::addChild(name, child);
        }
    }

    Float getRoughness(const Intersection &its, int component) const {
        return m_tex_alphas[0]->eval(its).average();
    }

    std::string toString() const {
       /*
        std::ostringstream oss;
        oss << "LayeredDielectric[" << endl
            << "  id = \"" << getID() << "\"," << endl
            << "  distribution = " << MicrofacetDistribution::distributionName(m_type) << "," << endl
            << "  sampleVisible = " << m_sampleVisible << "," << endl
            << "  alphaU = " << indent(m_alphaU->toString()) << "," << endl
            << "  alphaV = " << indent(m_alphaV->toString()) << "," << endl
            << "  specularReflectance = " << indent(m_specularReflectance->toString()) << "," << endl
            << "  eta = " << m_eta.toString() << "," << endl
            << "  k = " << m_k.toString() << endl
            << "]";
        return oss.str();
        */
        return std::string("");
    }

    MTS_DECLARE_CLASS()
private:

    /* Layered structure */
    int nb_layers;
    std::vector<Float>    m_depths;
    std::vector<Spectrum> m_sigma_a;
    std::vector<Spectrum> m_sigma_s;
    std::vector<Float>    m_gs;

    /* Associated maps */
	std::vector<ref<const Texture> > m_tex_etas;
    std::vector<ref<const Texture> > m_tex_kappas;
    std::vector<ref<const Texture> > m_tex_alphas;
    std::vector<ref<const Texture> > m_tex_depths;

    /* Precomputed data */
    FGD m_FGD;
    TIR m_TIR;
    bool m_useFGD;
    bool m_useTIR;
};

MTS_IMPLEMENT_CLASS_S(LayeredDielectric, false, BSDF)
MTS_EXPORT_PLUGIN(LayeredDielectric, "Rough conductor BRDF");
MTS_NAMESPACE_END
