// Mitsuba includes
#include <mitsuba/core/fresolver.h>
#include <mitsuba/core/statistics.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/hw/basicshader.h>

// STL includes
//#include <random>

// Local includes
#include "microfacet.h"
#include "ior.h"

// Local includes
// #include "brdf.h"
#include "layered.hpp"

MTS_NAMESPACE_BEGIN


/* Interact with a conducting layer
 *
 * Inputs:
 *    + `wi` input direction
 *    + `eta   =   eta2 / eta1` upper index of refaction
 *    + `kappa = kappa2 / eta1` complex index of refraction
 *    + `alpha` roughness of the interface
 *    + `uv` two random numbers in [0, 1]
 *
 * Outputs:
 *    + `wo` outgoing direction
 *    + `return value` reflectance of the layer
 */
Spectrum interactWithConductor(const Vector3& wi,
                               const Spectrum& eta,
                               const Spectrum& kappa,
                               float alpha,
                               const Vector2& uv,
                               Vector3& wo,
                               MicrofacetDistribution::EType type = MicrofacetDistribution::EGGX) {

   /* Construct the microfacet distribution matching the
      roughness values at the current surface position. */
   MicrofacetDistribution distr(type, alpha, alpha, true);

   /* Sample M, the microfacet normal */
   Float pdf; Point2 sample(uv.x, uv.y);
   Normal m = distr.sample(wi, sample, pdf);

   if (pdf == 0)
      return Spectrum(0.0f);

   /* Perfect specular reflection based on the microfacet normal */
   wo = reflect(wi, m);

   /* Side check */
   if (Frame::cosTheta(wo) <= 0)
      return Spectrum(0.0f);

   Spectrum F   = fresnelConductorExact(dot(wi, m), eta, kappa);
   //F = Spectrum(1.0f);
   Float weight = distr.smithG1(wo, m);

   return F * weight;
}


/* Interact with a dielectric layer
 *
 * Inputs:
 *    + `wi` input direction
 *    + `eta   =   eta2 / eta1` upper index of refaction
 *    + `kappa = kappa2 / eta1` complex index of refraction
 *    + `alpha` roughness of the interface
 *    + `uv` two random numbers in [0, 1]
 *
 * Outputs:
 *    + `wo` outgoing direction
 *    + `is_reflected` light flux direction
 *    + `return value` reflectance of the layer
 */
Spectrum interactWithDielectric(const Vector3& wi,
                                float eta,
                                float alpha,
                                const Vector3& uvw,
                                Vector3& wo,
                                bool& is_reflected,
                                MicrofacetDistribution::EType type = MicrofacetDistribution::EGGX) {

    /* Construct the microfacet distribution matching the
        roughness values at the current surface position. */
    MicrofacetDistribution distr(type, alpha, alpha, true);

    /* Sample M, the microfacet normal */
    Float microfacetPDF;
    Point2 sample(uvw.x, uvw.y);
    const Normal m = distr.sample(wi, sample, microfacetPDF);
    if (microfacetPDF == 0)
        return Spectrum(0.0f);

    Float cosThetaT;
    Float F = fresnelDielectricExt(dot(wi, m), cosThetaT, eta);
    Spectrum weight(1.0f);

    is_reflected = uvw.z <= F;

    if(is_reflected) {
      /* Perfect specular reflection based on the microfacet normal */
      wo = reflect(wi, m);

      /* Side check */
      if (Frame::cosTheta(wo) <= 0)
         return Spectrum(0.0f);

    } else {
        if (cosThetaT == 0)
            return Spectrum(0.0f);

        /* Perfect specular transmission based on the microfacet normal */
        wo = -refract(wi, m, eta, cosThetaT);

        /* Side check */
        if (Frame::cosTheta(wo) <= 0)
            return Spectrum(0.0f);

        //weight /= wo.z/wi.z;
   }

   weight *= distr.smithG1(wo, m);
   return weight;
}


class LayeredReference : public BSDF {
public:
    LayeredReference(const Properties &props) : BSDF(props), rng(new Random) {
        ref<FileResolver> fResolver = Thread::getThread()->getFileResolver();

        m_specularReflectance = new ConstantSpectrumTexture(
            props.getSpectrum("specularReflectance", Spectrum(1.0f)));

        std::string materialName = props.getString("material", "Cu");

        Spectrum intEta, intK;
        if (boost::to_lower_copy(materialName) == "none") {
            intEta = Spectrum(0.0f);
            intK = Spectrum(1.0f);
        } else {
            intEta.fromContinuousSpectrum(InterpolatedSpectrum(
                fResolver->resolve("data/ior/" + materialName + ".eta.spd")));
            intK.fromContinuousSpectrum(InterpolatedSpectrum(
                fResolver->resolve("data/ior/" + materialName + ".k.spd")));
        }

        Float extEta = lookupIOR(props, "extEta", "air");

        m_eta = props.getSpectrum("eta", intEta) / extEta;
        m_k   = props.getSpectrum("k", intK) / extEta;

        MicrofacetDistribution distr(props);
        m_type = distr.getType();
        m_sampleVisible = distr.getSampleVisible();

        m_alphaU = new ConstantFloatTexture(distr.getAlphaU());
        if (distr.getAlphaU() == distr.getAlphaV())
            m_alphaV = m_alphaU;
        else
            m_alphaV = new ConstantFloatTexture(distr.getAlphaV());

        // Global parameters for the layered amterials
        max_bounces = props.getInteger("max_bounces", 20);

        /* Parse a layered structure */
        parseLayers(props, nb_layers, m_etas, m_kappas, m_alphas,
                                      m_depths, m_sigma_s, m_sigma_a, m_gs);
    }

    LayeredReference(Stream *stream, InstanceManager *manager)
     : BSDF(stream, manager), rng(new Random) {
        m_type = (MicrofacetDistribution::EType) stream->readUInt();
        m_sampleVisible = stream->readBool();
        m_alphaU = static_cast<Texture *>(manager->getInstance(stream));
        m_alphaV = static_cast<Texture *>(manager->getInstance(stream));
        m_specularReflectance = static_cast<Texture *>(manager->getInstance(stream));
        m_eta = Spectrum(stream);
        m_k = Spectrum(stream);

        configure();
    }

    void serialize(Stream *stream, InstanceManager *manager) const {
        BSDF::serialize(stream, manager);

        stream->writeUInt((uint32_t) m_type);
        stream->writeBool(m_sampleVisible);
        manager->serialize(stream, m_alphaU.get());
        manager->serialize(stream, m_alphaV.get());
        manager->serialize(stream, m_specularReflectance.get());
        m_eta.serialize(stream);
        m_k.serialize(stream);
    }

    void configure() {
        unsigned int extraFlags = 0;
        if (m_alphaU != m_alphaV)
            extraFlags |= EAnisotropic;

        if (!m_alphaU->isConstant() || !m_alphaV->isConstant() ||
            !m_specularReflectance->isConstant())
            extraFlags |= ESpatiallyVarying;

        m_components.clear();
        m_components.push_back(EGlossyReflection | EFrontSide | extraFlags);

        /* Verify the input parameters and fix them if necessary */
        m_specularReflectance = ensureEnergyConservation(
            m_specularReflectance, "specularReflectance", 1.0f);

        m_usesRayDifferentials =
            m_alphaU->usesRayDifferentials() ||
            m_alphaV->usesRayDifferentials() ||
            m_specularReflectance->usesRayDifferentials();

        BSDF::configure();
    }

    // Random number generator
    ref<Random> rng;
    float generateRandomNumber() const {
        return const_cast<Random*>(rng.get())->nextFloat();
    }

    /// Helper function: reflect \c wi with respect to a given surface normal
    inline Vector reflect(const Vector &wi, const Normal &m) const {
        return 2 * dot(wi, m) * Vector(m) - wi;
    }

    /* WARNING: This code doesn't consider media scattering for the moment !
     */
    Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
        //return Spectrum(0.0f);
        /* Stop if this component was not requested */
        if (measure != ESolidAngle ||
            Frame::cosTheta(bRec.wi) <= 0 ||
            Frame::cosTheta(bRec.wo) <= 0 ||
            ((bRec.component != -1 && bRec.component != 0) ||
            !(bRec.typeMask & EGlossyReflection)))
            return Spectrum(0.0f);

      /* Energy of the photon. It starts at one and is decreased with
       * time. To have an infinite interaction loop, we need to perform
       * russian roulette on this value. */
        Vector3 wi   = bRec.wi;
        Vector3 wo   = bRec.wo;
        int index    = 0;
        int bounces  = 1;
      Spectrum e   = Spectrum(1.0);
      Spectrum f   = Spectrum(0.0);
      bool is_up   = false;

        /* Start the iteration loop */
        while(bounces <= max_bounces) {

            /* Obtain the material information */
            float    alpha  = m_alphas[index];
            Spectrum eta1   = (!is_up) ?   m_etas[index]   :   m_etas[index+1] ;
            Spectrum eta2   = (!is_up) ?   m_etas[index+1] :   m_etas[index]   ;
            Spectrum kappa2 = (!is_up) ? m_kappas[index+1] : m_kappas[index]   ;

         Spectrum kappa  = kappa2 / eta1;
         Spectrum eta    = eta2   / eta1;


            // Check if the photon is leaving the layer. This is the only
            // 'winning' situation where we have to record an outgoing direction.
            //
            if(index < 0 || index >= nb_layers) {
                break;

            // The conductor case on the first layer should resolve in a classic
            // rough conductor brdf model. So this conditional will only be applied
            // once and leave the function.
            //
            } else if(index == 0 && !is_up) {

            /* Calculate the reflection half-vector */
            Vector H = normalize(wo+wi);

            /* Construct the microfacet distribution matching the
               roughness values at the current surface position. */
            MicrofacetDistribution distr(m_type, alpha, alpha, true);

            /* Evaluate the microfacet normal distribution */
            const Float D = distr.eval(H);
            if (D == 0) {
               return Spectrum(0.0f);
            }

            /* Fresnel factor */
            const Spectrum F = fresnelConductorExact(dot(wi, H), eta, kappa);
            const Float    G = distr.G(wi, wo, H);

            /* Calculate the total amount of reflection */
            Float model = D * G / (4.0f * Frame::cosTheta(wi));

            f += e * F * model;

            } else if (index == 0) {

            Float n12 = eta.average();
            Vector3 wt = -wi;

            Vector H = normalize(wo*n12 + wt);
            H *= math::signum(Frame::cosTheta(H));

            /* Construct the microfacet distribution matching the
               roughness values at the current surface position. */
            MicrofacetDistribution distr(m_type, alpha, alpha, true);

            /* Evaluate the microfacet normal distribution */
            const Float D = distr.eval(H);
            if (D == 0)
               return Spectrum(0.0f);

            /* Fresnel factor */
            const Float F = fresnelDielectricExt(dot(wt, H), n12);

            /* Smith's shadow-masking function */
            const Float G = distr.G(wt, wo, H);

            /* Calculate the total amount of transmission */
            Float sqrtDenom = dot(wt, H) + n12 * dot(wo, H);
            Float value = ((1 - F) * D * G * dot(wt, H) * dot(wo, H)) / (Frame::cosTheta(wt) * sqrtDenom * sqrtDenom);

            /* Missing term in the original paper: account for the solid angle
               compression when tracing radiance -- this is necessary for
               bidirectional methods */
            //Float factor = (bRec.mode == ERadiance)
            //   ? (Frame::cosTheta(bRec.wi) > 0 ? m_invEta : m_eta) : 1.0f;

            f += std::abs(value) * e;
            }

            /* Get the next interaction with the layer */
         Vector3 wt;
         if(kappa.isZero()) {
               Vector3 uvw(generateRandomNumber(), generateRandomNumber(), generateRandomNumber());
            bool is_reflected = true;
               e *= interactWithDielectric(wi, eta.average(), alpha, uvw, wt, is_reflected, m_type);
            if(is_reflected) {
               is_up = !is_up;
            }

         } else {
               Vector2 uv(generateRandomNumber(), generateRandomNumber());
               e *= interactWithConductor(wi, eta, kappa, alpha, uv, wt, m_type);
            is_up = !is_up;
         }

         // /* Russian roulette */
         // float roulette = std::max<float>(e, 0.01);
         // if(roulette < dist(rng)) break;
            // e *= 1.0/(1.0 - roulette);
            if(e.isZero()) { break; }

            /* If the ray continues its travel, updates its parameters */
            wi       = wt;
            bounces += 1;
            index   += (is_up) ? -1 : 1;
        }

        if(!f.isValid()) {
            throw;
        }

        return f;
    }

    Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {
        if (measure != ESolidAngle ||
            Frame::cosTheta(bRec.wi) <= 0 ||
            Frame::cosTheta(bRec.wo) <= 0 ||
            ((bRec.component != -1 && bRec.component != 0) ||
            !(bRec.typeMask & EGlossyReflection)))
            return 0.0f;

        /* Calculate the reflection half-vector */
        Vector H = normalize(bRec.wo+bRec.wi);

        /* Construct the microfacet distribution matching the
           roughness values at the current surface position. */
        Float alpha = 0.0f;
        for(int index=0; index<nb_layers; ++index) {
            alpha = fmax(alpha, m_alphas[index]);
        }

        MicrofacetDistribution distr(m_type, alpha, alpha, true);

        // Evaluate the NDF
        const float D     = distr.eval(H);
        const float G     = distr.smithG1(bRec.wi, H);

        return 0.9*D*G / (4.0f * Frame::cosTheta(bRec.wi)) + 0.1;
    }

    Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
        Float pdf;
        return this->sample(bRec, pdf, sample);
    }

// #define MEDIA_MULTIPLE_SCATTERING

    Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf, const Point2 &sample) const {

        if (Frame::cosTheta(bRec.wi) < 0 ||
            ((bRec.component != -1 && bRec.component != 0) ||
            !(bRec.typeMask & EGlossyReflection)))
            return Spectrum(0.0f);

          /* Characteristics of the photon: direction, energy, position in the layered
         * structure. */
        Vector3 wi   = bRec.wi;
        int index    = 0;
        int bounces  = 0;
        Spectrum e   = Spectrum(1.0);
        bool is_up   = false;

        /* Start the iteration loop */
        while(bounces < max_bounces) {

            /* Retreive the layer depth. If set, we assume that the layer is a scattering medium
             * and sample the transmittance. We can as well do scattering in the medium
             */
            Float depth = m_depths[index];
            if(depth > 0.0f) {
                /* Obtain the meidum information */
                Spectrum sigma_a = m_sigma_a[index];
                Spectrum sigma_s = m_sigma_s[index];
                Spectrum sigma_t = (sigma_a+sigma_s);
                Float    sigma_f = sigma_t.average();
                Float    g       = m_gs[index];

                /* Remaining depth along the layer dimension */
                Float rmn_depth = depth;

#ifdef MEDIA_MULTIPLE_SCATTERING
                /* Perform multiple scattering in the layer */
                while(true) {
#endif
                    /* Sample the transmittance */
                    Float d  = - log(1.0 - bRec.sampler->next1D()) / sigma_f;
                    Float rd = d * Frame::cosTheta(wi); // Depth along the layer dimension

                    /* Should we stop the ray and sample the phase function? */
                    bool is_stopped   = rd < rmn_depth;
                    if(is_stopped) {

                        // Sample the HG function
                        Vector2 uv = Vector2(bRec.sampler->next1D(), bRec.sampler->next1D());
			               Vector3 wo = HG_Sample(wi, g, uv);

                        // Check if the new ray is up
                        bool is_wo_up = wo.z >= 0.0;
                        bool is_wi_up = wi.z >= 0.0;
                        bool chg_dir  = is_wo_up != is_wi_up;
                        if(chg_dir) {
                            wo = -wo;
#ifndef MEDIA_MULTIPLE_SCATTERING
                            return Spectrum(0.0f); // TODO remove: only for forward SS
#endif
                        }
                        if(wo.z == 0.0f) return Spectrum(0.0f);

                        // Sampling pdf
                        const float pdf = sigma_f * exp(-sigma_f *d);

                        // Evaluate ray's energy. If the ray change of direction of propagation
                        // use the opposite ray length.
                        e *= sigma_s*(- sigma_t * d).exp() / pdf;
                        Assert(e.isValid());

                        // Update transport information
                        bounces  += 1;
                        wi        = wo;
                        is_up     = (chg_dir) ? !is_up : is_up;
                        rmn_depth = (chg_dir) ? rd : rmn_depth - rd;

#ifndef MEDIA_MULTIPLE_SCATTERING
                        // Evaluate the end path attenuation
                        e *= (-sigma_t * rmn_depth/fabs(wi.z)).exp();
                        Assert(e.isValid());
#endif

                    /* Leave to the next interface */
                    } else {
                        const Float opt_depth = rmn_depth/Frame::cosTheta(wi);
                        if(std::isnan(opt_depth) || std::isinf(opt_depth)) { return Spectrum(0.0f); }

                        // Sampling pdf
                        const float pdf = exp(-sigma_f * opt_depth);

                        e *= (- sigma_t * opt_depth).exp() / pdf;
                        Assert(e.isValid());
#ifdef MEDIA_MULTIPLE_SCATTERING
                        break;
#endif
                    }
                    /* Short stop if the ray has no energy */
                    if(e.isZero()) { return Spectrum(0.0); }
                    Assert(e.isValid());
#ifdef MEDIA_MULTIPLE_SCATTERING
                }
#endif
            } else {
                /* Obtain the material information */
                float alpha     = m_alphas[index];
                Spectrum eta1   = (!is_up) ?   m_etas[index]   :   m_etas[index+1] ;
                Spectrum eta2   = (!is_up) ?   m_etas[index+1] :   m_etas[index]   ;
                Spectrum kappa2 = (!is_up) ? m_kappas[index+1] : m_kappas[index]   ;

                Spectrum eta    = eta2   / eta1;
                Spectrum kappa  = kappa2 / eta1;

                /* Get the next interaction with the layer */
                Vector3 wo;
                if(kappa.isZero()) {
                    Vector3 uvw(bRec.sampler->next1D(), bRec.sampler->next1D(), bRec.sampler->next1D());
                    //Vector3 uvw(generateRandomNumber(), generateRandomNumber(), generateRandomNumber());
                    bool is_reflected = true;
                    e *= interactWithDielectric(wi, eta.average(), alpha, uvw, wo, is_reflected, m_type);
                    if(is_reflected) {
                        is_up = !is_up;
                    }

                } else {
                    Vector2 uv(bRec.sampler->next1D(), bRec.sampler->next1D());
                    //Vector2 uv(generateRandomNumber(), generateRandomNumber());
                    e *= interactWithConductor(wi, eta, kappa, alpha, uv, wo, m_type);
                    is_up = !is_up;
                }

                // /* Russian roulette */
                // if(e.average() < dist(rng)) break;
                // e *= 1.0/(1.0 - e.average());
                if(e.isZero()) { return Spectrum(0.0f); }

                /* If the ray continues its travel, updates its parameters */
                wi       = wo;
                bounces += 1;
            }

            /* Update transport information */
            index   += (is_up) ? -1 : 1;

            // Check if the photon is leaving the layer. This is the only
            // 'winning' situation where we have to record an outgoing direction.
            if(index < 0) {

                bRec.wo = wi;
                bRec.eta = 1.0f;
                bRec.sampledComponent = 0;
                bRec.sampledType = EGlossyReflection;
                pdf = this->pdf(bRec, ESolidAngle);

                Assert(e.isValid());
                return e;

            } else if(index >= nb_layers) {
                return Spectrum(0.0f);
            }
        }

        return Spectrum(0.0f);
    }

    void addChild(const std::string &name, ConfigurableObject *child) {
        if (child->getClass()->derivesFrom(MTS_CLASS(Texture))) {
            if (name == "alpha")
                m_alphaU = m_alphaV = static_cast<Texture *>(child);
            else if (name == "alphaU")
                m_alphaU = static_cast<Texture *>(child);
            else if (name == "alphaV")
                m_alphaV = static_cast<Texture *>(child);
            else if (name == "specularReflectance")
                m_specularReflectance = static_cast<Texture *>(child);
            else
                BSDF::addChild(name, child);
        } else {
            BSDF::addChild(name, child);
        }
    }

    Float getRoughness(const Intersection &its, int component) const {
        return 0.5f * (m_alphaU->eval(its).average()
            + m_alphaV->eval(its).average());
    }

    std::string toString() const {
        std::ostringstream oss;
        oss << "LayeredReference[" << endl
            << "  id = \"" << getID() << "\"," << endl
            << "  distribution = " << MicrofacetDistribution::distributionName(m_type) << "," << endl
            << "  sampleVisible = " << m_sampleVisible << "," << endl
            << "  alphaU = " << indent(m_alphaU->toString()) << "," << endl
            << "  alphaV = " << indent(m_alphaV->toString()) << "," << endl
            << "  specularReflectance = " << indent(m_specularReflectance->toString()) << "," << endl
            << "  eta = " << m_eta.toString() << "," << endl
            << "  k = " << m_k.toString() << endl
            << "]";
        return oss.str();
    }

    Shader *createShader(Renderer *renderer) const;

    MTS_DECLARE_CLASS()
private:
    MicrofacetDistribution::EType m_type;
    ref<Texture> m_specularReflectance;
    ref<Texture> m_alphaU, m_alphaV;
    bool m_sampleVisible;
    Spectrum m_eta, m_k;

    /* Layered structure */
    int nb_layers;
    std::vector<Spectrum> m_etas;
    std::vector<Spectrum> m_kappas;
    std::vector<Float>    m_alphas;
    std::vector<Float>    m_depths;
    std::vector<Spectrum> m_sigma_a;
    std::vector<Spectrum> m_sigma_s;
    std::vector<Float>    m_gs;

    /* Constants */
    int max_bounces = 10;
};

/**
 * GLSL port of the rough conductor shader. This version is much more
 * approximate -- it only supports the Ashikhmin-Shirley distribution,
 * does everything in RGB, and it uses the Schlick approximation to the
 * Fresnel reflectance of conductors. When the roughness is lower than
 * \alpha < 0.2, the shader clamps it to 0.2 so that it will still perform
 * reasonably well in a VPL-based preview.
 */
class LayeredReferenceShader : public Shader {
public:
    LayeredReferenceShader(Renderer *renderer, const Texture *specularReflectance,
            const Texture *alphaU, const Texture *alphaV, const Spectrum &eta,
            const Spectrum &k) : Shader(renderer, EBSDFShader),
            m_specularReflectance(specularReflectance), m_alphaU(alphaU), m_alphaV(alphaV){
        m_specularReflectanceShader = renderer->registerShaderForResource(m_specularReflectance.get());
        m_alphaUShader = renderer->registerShaderForResource(m_alphaU.get());
        m_alphaVShader = renderer->registerShaderForResource(m_alphaV.get());

        /* Compute the reflectance at perpendicular incidence */
        m_R0 = fresnelConductorExact(1.0f, eta, k);
    }

    bool isComplete() const {
        return m_specularReflectanceShader.get() != NULL &&
               m_alphaUShader.get() != NULL &&
               m_alphaVShader.get() != NULL;
    }

    void putDependencies(std::vector<Shader *> &deps) {
        deps.push_back(m_specularReflectanceShader.get());
        deps.push_back(m_alphaUShader.get());
        deps.push_back(m_alphaVShader.get());
    }

    void cleanup(Renderer *renderer) {
        renderer->unregisterShaderForResource(m_specularReflectance.get());
        renderer->unregisterShaderForResource(m_alphaU.get());
        renderer->unregisterShaderForResource(m_alphaV.get());
    }

    void resolve(const GPUProgram *program, const std::string &evalName, std::vector<int> &parameterIDs) const {
        parameterIDs.push_back(program->getParameterID(evalName + "_R0", false));
    }

    void bind(GPUProgram *program, const std::vector<int> &parameterIDs, int &textureUnitOffset) const {
        program->setParameter(parameterIDs[0], m_R0);
    }

    void generateCode(std::ostringstream &oss,
            const std::string &evalName,
            const std::vector<std::string> &depNames) const {
        oss << "uniform vec3 " << evalName << "_R0;" << endl
            << endl
            << "float " << evalName << "_D(vec3 m, float alphaU, float alphaV) {" << endl
            << "    float ct = cosTheta(m), ds = 1-ct*ct;" << endl
            << "    if (ds <= 0.0)" << endl
            << "        return 0.0f;" << endl
            << "    alphaU = 2 / (alphaU * alphaU) - 2;" << endl
            << "    alphaV = 2 / (alphaV * alphaV) - 2;" << endl
            << "    float exponent = (alphaU*m.x*m.x + alphaV*m.y*m.y)/ds;" << endl
            << "    return sqrt((alphaU+2) * (alphaV+2)) * 0.15915 * pow(ct, exponent);" << endl
            << "}" << endl
            << endl
            << "float " << evalName << "_G(vec3 m, vec3 wi, vec3 wo) {" << endl
            << "    if ((dot(wi, m) * cosTheta(wi)) <= 0 || " << endl
            << "        (dot(wo, m) * cosTheta(wo)) <= 0)" << endl
            << "        return 0.0;" << endl
            << "    float nDotM = cosTheta(m);" << endl
            << "    return min(1.0, min(" << endl
            << "        abs(2 * nDotM * cosTheta(wo) / dot(wo, m))," << endl
            << "        abs(2 * nDotM * cosTheta(wi) / dot(wi, m))));" << endl
            << "}" << endl
            << endl
            << "vec3 " << evalName << "_schlick(float ct) {" << endl
            << "    float ctSqr = ct*ct, ct5 = ctSqr*ctSqr*ct;" << endl
            << "    return " << evalName << "_R0 + (vec3(1.0) - " << evalName << "_R0) * ct5;" << endl
            << "}" << endl
            << endl
            << "vec3 " << evalName << "(vec2 uv, vec3 wi, vec3 wo) {" << endl
            << "   if (cosTheta(wi) <= 0 || cosTheta(wo) <= 0)" << endl
            << "    	return vec3(0.0);" << endl
            << "   vec3 H = normalize(wi + wo);" << endl
            << "   vec3 reflectance = " << depNames[0] << "(uv);" << endl
            << "   float alphaU = max(0.2, " << depNames[1] << "(uv).r);" << endl
            << "   float alphaV = max(0.2, " << depNames[2] << "(uv).r);" << endl
            << "   float D = " << evalName << "_D(H, alphaU, alphaV)" << ";" << endl
            << "   float G = " << evalName << "_G(H, wi, wo);" << endl
            << "   vec3 F = " << evalName << "_schlick(1-dot(wi, H));" << endl
            << "   return reflectance * F * (D * G / (4*cosTheta(wi)));" << endl
            << "}" << endl
            << endl
            << "vec3 " << evalName << "_diffuse(vec2 uv, vec3 wi, vec3 wo) {" << endl
            << "    if (cosTheta(wi) < 0.0 || cosTheta(wo) < 0.0)" << endl
            << "    	return vec3(0.0);" << endl
            << "    return " << evalName << "_R0 * inv_pi * inv_pi * cosTheta(wo);"<< endl
            << "}" << endl;
    }
    MTS_DECLARE_CLASS()
private:
    ref<const Texture> m_specularReflectance;
    ref<const Texture> m_alphaU;
    ref<const Texture> m_alphaV;
    ref<Shader> m_specularReflectanceShader;
    ref<Shader> m_alphaUShader;
    ref<Shader> m_alphaVShader;
    Spectrum m_R0;
};

Shader *LayeredReference::createShader(Renderer *renderer) const {
    return new LayeredReferenceShader(renderer,
        m_specularReflectance.get(), m_alphaU.get(), m_alphaV.get(), m_eta, m_k);
}

MTS_IMPLEMENT_CLASS(LayeredReferenceShader, false, Shader)
MTS_IMPLEMENT_CLASS_S(LayeredReference, false, BSDF)
MTS_EXPORT_PLUGIN(LayeredReference, "Rough conductor BRDF");
MTS_NAMESPACE_END
