// Mitsuba includes
#include <mitsuba/core/fresolver.h>
#include <mitsuba/render/bsdf.h>
#include <mitsuba/hw/basicshader.h>

// STL includes
//#include <random>

// Local includes
#include "microfacet.h"
#include "ior.h"

// Local includes
#include "layered.hpp"


MTS_NAMESPACE_BEGIN

class LayeredSymmetric : public BSDF {
public:
    LayeredSymmetric(const Properties &props) : BSDF(props) {
        ref<FileResolver> fResolver = Thread::getThread()->getFileResolver();

        /* Parse a layered structure */
        parseLayers(props, nb_layers, m_tex_etas, m_tex_kappas, m_tex_alphas,
                                      m_depths, m_sigma_s, m_sigma_a, m_gs);

        // Load the FGD and TIR textures
        m_useFGD = props.getBoolean("useFGD", false);
        if(m_useFGD) {
            m_FGD = FGD(fResolver->resolve("data/FGD.bin").string());
        }
        m_useTIR = props.getBoolean("useTIR", false);
        if(m_useTIR){
            m_TIR = TIR(fResolver->resolve("data/TIR.bin").string());
        }
    }

    LayeredSymmetric(Stream *stream, InstanceManager *manager)
     : BSDF(stream, manager) {
        /*
        m_type = (MicrofacetDistribution::EType) stream->readUInt();
        m_sampleVisible = stream->readBool();
        m_alphaU = static_cast<Texture *>(manager->getInstance(stream));
        m_alphaV = static_cast<Texture *>(manager->getInstance(stream));
        m_specularReflectance = static_cast<Texture *>(manager->getInstance(stream));
        m_eta = Spectrum(stream);
        m_k = Spectrum(stream);
         */
        configure();
    }

    void serialize(Stream *stream, InstanceManager *manager) const {
        BSDF::serialize(stream, manager);

        /*
        stream->writeUInt((uint32_t) m_type);
        stream->writeBool(m_sampleVisible);
        manager->serialize(stream, m_alphaU.get());
        manager->serialize(stream, m_alphaV.get());
        manager->serialize(stream, m_specularReflectance.get());
        m_eta.serialize(stream);
        m_k.serialize(stream);
        */
    }

    void configure() {
        unsigned int extraFlags = 0;
        /*
        if (!m_alphaU->isConstant() || !m_alphaV->isConstant() ||
            !m_specularReflectance->isConstant())
            extraFlags |= ESpatiallyVarying;
         */

        m_components.clear();
        m_components.push_back(EGlossyReflection | EFrontSide | extraFlags);

        m_usesRayDifferentials = false;
        /*
            m_alphaU->usesRayDifferentials() ||
            m_alphaV->usesRayDifferentials() ||
            m_specularReflectance->usesRayDifferentials();
         */

        BSDF::configure();
    }

    /*
     */
    void evalFresnel(Float ct, Float alpha, const Spectrum& eta, const Spectrum& kappa,
                     Spectrum& Rij, Spectrum& Tij) const {

        if(m_useFGD) {
            Rij = m_FGD(ct, alpha, eta, kappa);
        } else {
            Rij = (kappa.isZero()) ? fresnelDielectricExt(ct, eta[0]) * Spectrum(1.0f) :
                                     fresnelConductorExact(ct, eta, kappa);
        }
        Tij = (kappa.isZero()) ? Spectrum(1.0) - Rij : Spectrum(0.0);
    }

    /* Trace moment analysis in the layer structure for a given input angle
     */
    void computeAddingDoubling(Float _cti, const Spectrum* m_etas, const Spectrum* m_kappas, const Float* m_alphas,
                               Spectrum* coeffs, Float* alphas, int& nb_valid) const {

        // Variables
        float cti  = _cti;
        Spectrum R0i(0.0f), Ri0(0.0f), T0i(1.0f), Ti0(1.0f);
        float s_r0i = 0.0f, s_ri0=0.0f, s_t0i=0.0f, s_ti0=0.0f;
        float j0i=1.0f, ji0=1.0f;

        // Iterate over the layers
        for(int i=0; i<nb_layers; ++i) {

            /* Extract layer data */
            Spectrum eta_1   = m_etas[i];
            Spectrum eta_2   = m_etas[i+1];
            Spectrum kappa_2 = m_kappas[i+1];
            Spectrum eta     = eta_2 / eta_1;
            Spectrum kappa   = kappa_2 / eta_1;
            float alpha      = m_alphas[i];
            float n12        = eta.average();
            float depth      = m_depths[i];

            Spectrum R12, T12, R21, T21;
            float s_r12=0.0f, s_r21=0.0f, s_t12=0.0f, s_t21=0.0f, j12=1.0f, j21=1.0f, ctt;
            if(depth > 0.0f) {
                /* Mean doesn't change with volumes */
                ctt = cti;

                /* Evaluate transmittance */
                const Spectrum sigma_t = m_sigma_a[i] + m_sigma_s[i];
                const float ray_length = depth/ctt;
                T12 = (Spectrum(1.0f) + m_sigma_s[i]*ray_length) * (- ray_length * sigma_t).exp();
                T21 = T12;
                R12 = Spectrum(0.0f);
                R21 = Spectrum(0.0f);

                /* Fetch precomputed variance for HG phase function */
                s_t12 = alpha;
                s_t21 = alpha;

            } else {
                /* Evaluate off-specular transmission */
                float sti = sqrt(1.0f - cti*cti);
                float stt = sti / n12;
                if(stt <= 1.0f) {
                    //const float scale = _clamp<float>((1.0f-alpha)*(sqrt(1.0f-alpha) + alpha), 0.0f, 1.0f);
                    //stt = scale*stt + (1.0f-scale)*sti;
                    ctt = sqrt(1.0f - stt*stt);
                } else {
                    ctt = -1.0f;
                }

                /* Ray is not block by conducting interface or total reflection */
                const bool has_transmissive = ctt > 0.0f && kappa.isZero();

                /* Evaluate interface variance term */
                s_r12 = roughnessToVariance(alpha);
                s_r21 = s_r12;

                /* For dielectric interfaces, evaluate the transmissive roughnesses */
                if(has_transmissive) {
                   const float _ctt = 1.0f; // The scaling factor overblurs the BSDF at grazing
                   const float _cti = 1.0f; // angles (we cannot account for the deformation of
                                            // the lobe for those configurations.

                    s_t12 = roughnessToVariance(alpha * 0.5f * fabs(_ctt*n12 - _cti)/(_ctt*n12));
                    s_t21 = roughnessToVariance(alpha * 0.5f * fabs(_cti/n12 - _ctt)/(_cti/n12));
                      j12 = (ctt/cti) * n12; // Scale due to the interface
                      j21 = (cti/ctt) / n12;
                }

                /* Evaluate FGD using a modified roughness accounting for top layers */
                auto temp_alpha = varianceToRoughness(s_t0i + s_r12);

                /* Evaluate r12, r21, t12, t21 */
                evalFresnel(cti, temp_alpha, eta, kappa, R12, T12);
                if(has_transmissive) {
                    R21 = R12;
                    T21 = T12 /* (n12*n12)*/; // No need to account for IOR scaling since
                    T12 = T12 /* (n12*n12)*/; // The plugin is used for reflection only.
                } else {
                    R21 = Spectrum(0.0f);
                    T21 = Spectrum(0.0f);
                    T12 = Spectrum(0.0f);
                }

                /* Evaluate TIR using the decoupling approximation */
                if(i > 0 && m_useTIR) {
                    Spectrum eta_0   = (i==0) ? m_etas[0] : m_etas[i-1];
                    float n10        = (eta_0/eta_1).average();

                    const float _TIR  = m_TIR(cti, temp_alpha, n10);
                    Ri0 += (1.0f-_TIR) * Ti0;
                    for(int llid=0; llid<3; ++llid) { Ri0[llid] = fmin(Ri0[llid], 1.0f); }
                    Ti0 *= _TIR;
                }
            }


            /* Evaluate the adding operators on the energy */
            const Spectrum denom = (Spectrum(1.0f) - Ri0*R12);

            /* Multiple scattering forms */
            const float denom_avg = denom.average();
            const Spectrum m_R0i = (denom_avg <= 0.0f) ? Spectrum(0.0f) : (T0i*R12*Ti0) / denom;
            const Spectrum m_Ri0 = (denom_avg <= 0.0f) ? Spectrum(0.0f) : (T21*Ri0*T12) / denom;
            const Spectrum m_Rr  = (denom_avg <= 0.0f) ? Spectrum(0.0f) : (Ri0*R12) / denom;
            const float m_r0i    = m_R0i.average();
            const float m_ri0    = m_Ri0.average();
            const float m_rr     = m_Rr.average();

            /* Evaluate the adding operator on the energy */
            const Spectrum e_R0i = R0i + m_R0i;
            const Spectrum e_T0i = (T0i*T12) / denom;
            const Spectrum e_Ri0 = R21 + m_Ri0;
            const Spectrum e_Ti0 = (T21*Ti0) / denom;

            /* Scalar energy forms to compute adding on variance */
            const float r21   = R21.average();
            const float r0i   = R0i.average();
            const float e_r0i = e_R0i.average();
            const float e_ri0 = e_Ri0.average();

            /* Evaluate the adding operator on the normalized variance */
            float _s_r0i = (r0i*s_r0i + m_r0i*(s_ti0 + j0i*(s_t0i + s_r12 + m_rr*(s_r12+s_ri0))));// e_r0i;
            float _s_t0i = j12*s_t0i + s_t12 + j12*(s_r12 + s_ri0)*m_rr;
            float _s_ri0 = (r21*s_r21 + m_ri0*(s_t12 + j12*(s_t21 + s_ri0 + m_rr*(s_r12+s_ri0))));// e_ri0;
            float _s_ti0 = ji0*s_t21 + s_ti0 + ji0*(s_r12 + s_ri0)*m_rr;
            _s_r0i = (e_r0i > 0.0f) ? _s_r0i/e_r0i : 0.0f;
            _s_ri0 = (e_ri0 > 0.0f) ? _s_ri0/e_ri0 : 0.0f;

            /* Store the coefficient and variance */
            if(m_r0i > 0.0f) {
                coeffs[i] = m_R0i;
                alphas[i] = varianceToRoughness(s_ti0 + j0i*(s_t0i + s_r12 + m_rr*(s_r12+s_ri0)));
            } else {
                coeffs[i] = Spectrum(0.0f);
                alphas[i] = 0.0f;
            }

            /* Update energy */
            R0i = e_R0i;
            T0i = e_T0i;
            Ri0 = e_Ri0;
            Ti0 = e_Ti0;

            /* Update mean */
            cti = ctt;

            /* Update variance */
            s_r0i = _s_r0i;
            s_t0i = _s_t0i;
            s_ri0 = _s_ri0;
            s_ti0 = _s_ti0;

            /* Update jacobian */
            j0i *= j12;
            ji0 *= j21;

            /* Escape if a conductor is present */
            if(kappa.average() > 0.0f) {
               nb_valid = i+1;
               return;
            }
        }

        nb_valid = nb_layers;
    }

    /* Evaluate the maps */
    inline void evalMaps(const BSDFSamplingRecord &bRec,
                         Spectrum* m_etas, Spectrum* m_kappas,
                         Float* m_alphas) const {

        // Fetch air values
        m_etas[0]   = m_tex_etas[0]->eval(bRec.its);
        m_kappas[0] = m_tex_kappas[0]->eval(bRec.its);

        // Fetch textured values
        for(int i=0; i<nb_layers; ++i) {
            m_etas[i+1]   = m_tex_etas[i+1]->eval(bRec.its);
            m_kappas[i+1] = m_tex_kappas[i+1]->eval(bRec.its);
            m_alphas[i]   = m_tex_alphas[i]->eval(bRec.its).average();
            // TODO Add other maps
        }
    }

    Spectrum eval(const Vector3& wi, const Vector3& wo,
                  const Spectrum* m_etas, const Spectrum* m_kappas, const Float* m_alphas) const {

        /* Constants */
        Vector3 H = normalize(wo+wi);

        /* Result */
        Spectrum f(0.0f);

        /* Evaluate the adding method to get coeffs and variances */
        Spectrum coeffs[nb_layers];
        float    alphas[nb_layers];
        int nb_valid = 0;
        computeAddingDoubling(dot(wi, H), m_etas, m_kappas, m_alphas,
                              coeffs, alphas, nb_valid);

        /* Sum the contribution of all the interfaces */
        for(int index=0; index<nb_valid; ++index) {

            // Skip zero contributions
            if(coeffs[index].isZero()) {
                continue;
            }

            // Fetch current roughness
            const Float a = alphas[index];

            // Evaluate microfacet model
            MicrofacetDistribution distr(MicrofacetDistribution::EGGX, a, a, true);
            const Float D = distr.eval(H);
            const Float G = distr.G(wi, wo, H);

            // Add to the contribution
            f += D*G * coeffs[index] / (4.0f * Frame::cosTheta(wi));
        }
        return f;
    }

    Spectrum eval(const BSDFSamplingRecord &bRec, EMeasure measure) const {
        /* Stop if this component was not requested */
        if (measure != ESolidAngle ||
            Frame::cosTheta(bRec.wi) <= 0 ||
            Frame::cosTheta(bRec.wo) <= 0 ||
            ((bRec.component != -1 && bRec.component != 0) ||
            !(bRec.typeMask & EGlossyReflection)))
            return Spectrum(0.0f);

        /* Evaluate the maps */
        Float m_alphas[nb_layers];
        Spectrum m_kappas[nb_layers+1], m_etas[nb_layers+1];
        evalMaps(bRec, m_etas, m_kappas, m_alphas);

        auto result = this->eval(bRec.wi, bRec.wo, m_etas, m_kappas, m_alphas);
        Assert(!result.isNaN());
        return result;
    }

    Float pdf(const BSDFSamplingRecord &bRec, EMeasure measure) const {
        if (measure != ESolidAngle ||
            Frame::cosTheta(bRec.wi) <= 0 ||
            Frame::cosTheta(bRec.wo) <= 0 ||
            ((bRec.component != -1 && bRec.component != 0) ||
            !(bRec.typeMask & EGlossyReflection)))
            return 0.0f;

        /* Calculate the reflection half-vector */
        Vector H = normalize(bRec.wo+bRec.wi);

        /* Evaluate the maps */
        Float m_alphas[nb_layers];
        Spectrum m_kappas[nb_layers+1], m_etas[nb_layers+1];
        evalMaps(bRec, m_etas, m_kappas, m_alphas);

        /* Evaluate the adding method to get coeffs and variances */
        Spectrum coeffs[nb_layers];
        float    alphas[nb_layers];
        int nb_valid = 0;
        computeAddingDoubling(bRec.wi.z, m_etas, m_kappas, m_alphas,
                              coeffs, alphas, nb_valid);

        /* Convert Spectral coefficients to float for pdf weighting */
        float pdf   = 0.0;
        float cum_w = 0.0;
        for(int i=0; i<nb_valid; ++i) {
            // Skip zero contributions
            if(coeffs[i].isZero()) {
                continue;
            }

            /* Evlaluate weight */
            auto weight = coeffs[i].average();
            cum_w += weight;

            /* Generate NDF */
            MicrofacetDistribution distr(MicrofacetDistribution::EGGX,
                                        alphas[i], alphas[i], true);

            /* Evaluate the pdf */
            float DG = distr.eval(H) * distr.smithG1(bRec.wo, H) / (4.0f * Frame::cosTheta(bRec.wi));
            pdf += weight * DG;
        }

         if(cum_w > 0.0f) {
            return pdf / cum_w;
         } else {
            return 0.0f;
         }
    }

    Spectrum sample(BSDFSamplingRecord &bRec, const Point2 &sample) const {
        Float pdf;
        return this->sample(bRec, pdf, sample);
    }

    Spectrum sample(BSDFSamplingRecord &bRec, Float &pdf, const Point2 &sample) const {

        if (Frame::cosTheta(bRec.wi) < 0 ||
        ((bRec.component != -1 && bRec.component != 0) ||
        !(bRec.typeMask & EGlossyReflection)))
        return Spectrum(0.0f);

        Spectrum R;

        /* Evaluate the maps */
        Float m_alphas[nb_layers];
        Spectrum m_kappas[nb_layers+1], m_etas[nb_layers+1];
        evalMaps(bRec, m_etas, m_kappas, m_alphas);

        /* Evaluate the adding method to get coeffs and variances */
        Spectrum coeffs[nb_layers];
        float    alphas[nb_layers];
        int nb_valid = 0;
        computeAddingDoubling(bRec.wi.z, m_etas, m_kappas, m_alphas,
                              coeffs, alphas, nb_valid);

        /* Convert Spectral coefficients to floats to select BRDF lobe to sample */
        float weights[nb_valid];
        float cum_w = 0.0;
        for(int i=0; i<nb_valid; ++i) {
            weights[i] = coeffs[i].average();
            cum_w += weights[i];
        }

        /* Select a random BRDF lobe */
        float sel_w = bRec.sampler->next1D() * cum_w - weights[0];
        int   sel_i = 0;
        for(sel_i=0; sel_w>0.0 && sel_i<nb_valid; sel_i++) {
            sel_w -= weights[sel_i+1];
        }

        /* Sample a microfacet normal */
        MicrofacetDistribution distr(MicrofacetDistribution::EGGX,
                                     alphas[sel_i], alphas[sel_i], true);
        Normal m = distr.sample(bRec.wi, sample, pdf);


        /* Perfect specular reflection based on the microfacet normal */
        bRec.wo = reflect(bRec.wi, m);
        if(bRec.wo.z <= 0.0f || pdf <= 0.0f) {
            return Spectrum(0.0f);
        }

        bRec.eta = 1.0f;
        bRec.sampledComponent = 0;
        bRec.sampledType = EGlossyReflection;

        /* Evaluate the MIS 'pdf' using the balance heuristic */
        pdf = 0.0;
        for(int i=0; i<nb_valid; ++i) {
            MicrofacetDistribution distr(MicrofacetDistribution::EGGX, alphas[i], alphas[i], true);
            float DG = distr.eval(m) * distr.smithG1(bRec.wo, m) / (4.0f * Frame::cosTheta(bRec.wi));
            pdf += (weights[i]/cum_w) * DG;
        }

        /* Evaluate the BRDF and the final weight */
        R = this->eval(bRec.wi, bRec.wo, m_etas, m_kappas, m_alphas);
        Assert(!R.isNaN());

        if(pdf > 0.0f) {
         return R / pdf;
        } else {
           return Spectrum(0.0f);
        }
    }

    void addChild(const std::string &name, ConfigurableObject *child) {
        if (child->getClass()->derivesFrom(MTS_CLASS(Texture))) {
            std::string substr = name.substr(0, 3);
            if (substr == "eta") {
                int eta_index = atoi(name.substr(4).c_str())+1;
                std::cout << name << " => " << eta_index << " / " << nb_layers << std::endl;
                m_tex_etas[eta_index] = static_cast<Texture *>(child);
            } else if (substr == "kap") {
                int kappa_index = atoi(name.substr(6).c_str())+1;
                std::cout << name << " => " << kappa_index << " / " << nb_layers << std::endl;
                m_tex_kappas[kappa_index] = static_cast<Texture *>(child);
            }  else if (substr == "alp") {
                int alpha_index = atoi(name.substr(6).c_str());
                std::cout << name << " => " << alpha_index << " / " << nb_layers-1 << std::endl;
                m_tex_alphas[alpha_index] = static_cast<Texture *>(child);
            } else {
                // BSDF::addChild(name, child);
            }
        } else {
            BSDF::addChild(name, child);
        }
    }

    Float getRoughness(const Intersection &its, int component) const {
        return m_tex_alphas[0]->eval(its).average();
    }

    std::string toString() const {
       /*
        std::ostringstream oss;
        oss << "LayeredSymmetric[" << endl
            << "  id = \"" << getID() << "\"," << endl
            << "  distribution = " << MicrofacetDistribution::distributionName(m_type) << "," << endl
            << "  sampleVisible = " << m_sampleVisible << "," << endl
            << "  alphaU = " << indent(m_alphaU->toString()) << "," << endl
            << "  alphaV = " << indent(m_alphaV->toString()) << "," << endl
            << "  specularReflectance = " << indent(m_specularReflectance->toString()) << "," << endl
            << "  eta = " << m_eta.toString() << "," << endl
            << "  k = " << m_k.toString() << endl
            << "]";
        return oss.str();
        */
        return std::string("");
    }

    MTS_DECLARE_CLASS()
private:

    /* Layered structure */
    int nb_layers;
    std::vector<Float>    m_depths;
    std::vector<Spectrum> m_sigma_a;
    std::vector<Spectrum> m_sigma_s;
    std::vector<Float>    m_gs;

    /* Associated maps */
	std::vector<ref<const Texture> > m_tex_etas;
    std::vector<ref<const Texture> > m_tex_kappas;
    std::vector<ref<const Texture> > m_tex_alphas;
    std::vector<ref<const Texture> > m_tex_depths;

    /* Precomputed data */
    FGD m_FGD;
    TIR m_TIR;
    bool m_useFGD;
    bool m_useTIR;
};

MTS_IMPLEMENT_CLASS_S(LayeredSymmetric, false, BSDF)
MTS_EXPORT_PLUGIN(LayeredSymmetric, "Symmetric Layered BRDF");
MTS_NAMESPACE_END
