#line 1
precision highp float;

#include "shaders/library/common.shader"
#include "shaders/library/random.shader"
#include "shaders/library/brdf.shader"
#include "shaders/library/layer.shader"
#include "shaders/library/covariance.shader"
#line 9

// Uniform variables
uniform vec3  wi;
uniform float std_dev;

uniform float  u_Scale;
uniform float  u_Gamma;
uniform float  u_Depth;
uniform int    u_VizuSelect;
uniform int    u_ScalingSelect;

/* Gamma transformation
 */
vec3 Gamma(vec3 R) {
    return exp(u_Gamma * log(u_Scale * R));
}

/* Evaluate an emissive Gaussian term
 */
vec3 Emission(vec3 w, vec3 wc, float dev) {
    vec3 n = wc; n.xy = - n.xy;
    return TangentialGaussian(w, n, dev) * vec3(1,1,1);
}

void main(void) {

    vec3 N = vec3(0,0,1);

    // Get the omega_o direction from the vPos variable
    vec3 wo, wc, wr;
    wc    = wi;
    wo.xy = vPos.xy;
    float sinTo = length(wo.xy);
    if(sinTo <= 1.0) {
        float cosTo = sqrt(1.0 - sinTo*sinTo);
        wo.z = cosTo;

        /* Statistics for the absorbance
         */
        // float abs_dev = u_Depth;    // (1) This factor is given by the Hessian / energy.
        float abs_dev = sqrt(u_Depth); // (2) This factor works better, probably due 
                                       // to the shape of the GGX distribution.

        /* Convert to the correct variance space
         */
        float new_dev = roughnessToVariance(std_dev);                                       

        /* Modify the mean
         */
        // wc.xy = wc.xy * abs_dev / (new_dev+abs_dev); // (1) Theoretical mean using the 
                                                     // product of Gaussians
        wc.xy = wc.xy * exp(-sqr(std_dev)*u_Depth); // (2) Empirical mean
        wc.z  = sqrt(1.0 - sqr(length(wc.xy)));
        wr = reflect(wc, N);

        /* Modify the energy
         */
        float sinTi = length(wc.xy);
        float cosTi = wc.z;
        float absorption = exp(- u_Depth * sqrt(1.0 + sqr(sinTi/cosTi)));
        if(u_VizuSelect >= 1) {
            absorption /= exp(- u_Depth);
        }

        /* Modify the variance
         */

        if(u_ScalingSelect >= 1) {
            new_dev = new_dev / (abs_dev * (new_dev + 1.0/abs_dev));
        }
        new_dev = varianceToRoughness(new_dev);

        /* Evaluate the input GGX lobe */
        float GGX_r;
        GGX_r  = GGX(wc, wo, N, new_dev);
        GGX_r *= GGX_D(1.0, std_dev)/GGX_D(1.0, new_dev); // We need to scale the new
                                                          // NDF lobe by the old one
                                                          // to have the correct norm.


        /* Perform random evaluation of the layered structure using 'wiT' and 'woT'
         * as the incoming and outgoing directions.
         */
        if(wo.z == 0.0) {
            gl_FragColor.xyz = vec3(0.0, 0.0, 0.0);
        } else {
            gl_FragColor.xyz = Gamma(absorption * GGX_r * vec3(1,1,1));
            if(u_VizuSelect >= 2) {
                gl_FragColor.xyz = mod(10.*gl_FragColor.xyz, 1.0);
            }
        }
        gl_FragColor.w   = 1.0;

    } else {    
        gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);
    }
}
