#line 1
precision highp float;

#include "shaders/library/common.shader"
#include "shaders/library/brdf.shader"
#include "shaders/library/layer.shader"
#include "shaders/library/pivot.shader"
#line 8

// Uniform variables
uniform vec3   wi;
uniform int    u_ApproxMethod;
uniform float  u_Scale;
uniform float  u_Gamma;

/* Gamma transformation
 */
vec3 Gamma(vec3 R) {
    return exp(u_Gamma * log(u_Scale * R));
}
const float M_PI = 3.14159265358979323846;

void ComputeApproxRoughness(in vec3 wi, in vec3 wo, out float a, out float f) {

    // Set the initial values for the importance distribution
    a = 0.0;
    f = 1.0;

    // Temp variables
    float R12;
    vec3  wt, w, wto;

    // IOR ratio
    float eta1 = layers[0].n;
    float eta2 = layers[1].n;
    float ieta = eta1 / eta2;
    float eta  = eta2 / eta1;

    // Evaluate the Fresnel term for the transmitted rays
    Fresnel(wi, layers[0].n, layers[1].n, wt, R12);
    f *= abs(1.0 - R12);

    // First interface (between air [0] and first layer [1])
    // Scale the roughness for the transmission
    float r    = layers[1].a * 0.5*(eta*wt.z - wi.z)/(eta*wt.z);
    float s12  = sqr(r) / (2.0 - 2.0*sqr(r));
    a   += s12;

    // Convert back to roughness
    a  = sqrt(2.0 / ((1.0/a) + 2.0));
}

void ComputeRoughnessKulla(in vec3 wi, in vec3 wo, out float a, out float f) {

    // Set the initial values for the importance distribution
    a = 0.0;
    f = 1.0;

    // Temp variables
    float R12;
    vec3  wt, w, wto;

    // IOR ratio
    float eta1 = layers[0].n;
    float eta2 = layers[1].n;
    float ieta = eta1 / eta2;
    float eta  = eta2 / eta1;

    // Evaluate the Fresnel term for the transmitted rays
    Fresnel(wi, layers[0].n, layers[1].n, wt, R12);
    f *= abs(1.0 - R12);

    // Scale the roughness for the transmission
    a = sqrt(0.5*(eta - 1.0) / eta) * layers[1].a; // Using Kulla & Conty [2017] slide 39
}

void ComputeApproxVMF(in  vec3 wi, in  vec3 wo,
                      out float mu, out float km) {

    mu = 1.0;    
    km = 0.0;
    
    // Temp variables
    float R12;
    vec3  wt, w, wto;
    float km1, km2;

    // IOR ratio
    float eta1 = layers[0].n;
    float eta2 = layers[1].n;
    float ieta = eta1 / eta2;
    float eta  = eta2 / eta1;

    // Evaluate the Fresnel term for the transmitted rays
    Fresnel(wi, layers[0].n, layers[1].n, wt, R12);
    mu *= abs(1.0 - R12);

    // Exponent of the layer
    km1  = vMF_RoughnessToKm(layers[1].a);
    km1 *= sqr(eta) * (abs(wt.z)) / sqr(wo.z + eta*wt.z);
    
    // Exponent of the incident source. Using almost infinite
    // value to fake a dirac.
    km2  = 1.0E8;
    km2 *= sqr(eta) * abs(wt.z / wo.z);

    km = vMF_invA3(vMF_A3(km1) * vMF_A3(km2));
}

void main(void) {

    // Get the omega_o direction from the vPos variable
    vec3 wo;
    wo.xy = -vPos.xy;
    float sinTo = length(wo.xy);
    if(sinTo <= 1.0) {
        wo.z = sqrt(1.0 - sinTo*sinTo);
        vec3 wr = -reflect(wi, vec3(0.0, 0.0, 1.0));
        vec3 wt = SnellRefraction(wi, vec3(0.0, 0.0, 1.0), layers[0].n/layers[1].n);

        // Update the mean direction to match the off-specular direction
        float smoothness = 1.0 - layers[1].a;
        float lerpFactor = smoothness * (sqrt(smoothness) + layers[1].a);
        vec3  wa = lerpFactor*wt + (1.0-lerpFactor)*wi;
        wa.z = -sqrt(1.0 - (wa.x*wa.x + wa.y+wa.y));

        vec3 R = vec3(0.0, 0.0, 0.0);

        // Using covariance approximation for the resulting BRDF function.
        if(u_ApproxMethod == 0) {
            float a, f;
            ComputeApproxRoughness(wi, wo, a, f);

            float GGX_r;
            GGX_r  = GGX(-wa, wo, vec3(0.0, 0.0, 1.0), max(a, 1.0E-5));
            R     += f * GGX_r * vec3(1.0, 1.0, 1.0);

        } else if(u_ApproxMethod == 1) {
            float a, f;
            ComputeApproxRoughness(wi, wo, a, f);

            float GGX_r;
            GGX_r  = GGX(-wt, wo, vec3(0.0, 0.0, 1.0), max(a, 1.0E-5));
            R     += f * GGX_r * vec3(1.0, 1.0, 1.0);

        } else if(u_ApproxMethod == 2) {
            float mu, km;
            ComputeApproxVMF(wi, wo, mu, km);

            float vMF_r;
            vMF_r  = vMF_BRDF(-wt, wo, km);
            R     += mu * vMF_r * vec3(1.0, 1.0, 1.0);

        } else {
            float a, f;
            ComputeRoughnessKulla(wi, wo, a, f);

            float GGX_r;
            GGX_r  = GGX(-wt, wo, vec3(0.0, 0.0, 1.0), max(a, 1.0E-5));
            R     += f * GGX_r * vec3(1.0, 1.0, 1.0);
        }

        gl_FragColor.xyz = Gamma(R);
        gl_FragColor.w   = 1.0;

        // Display helpers that show the incoming ray and the purely reflected
        // ray using circles.
        float dist;
        dist = length(wo.xy - wt.xy);
        if(dist < 0.01) {
            gl_FragColor.xyz = vec3(0.0, 0.0, 1.0);
        }
        dist = length(wo.xy - wr.xy);
        if(dist < 0.01) {
            gl_FragColor.xyz = vec3(0.0, 1.0, 0.0);
        }
        dist = length(wo.xy - wi.xy);
        if(dist < 0.01) {
            gl_FragColor.xyz = vec3(1.0, 0.0, 0.0);
        }
        dist = length(wo.xy - wa.xy);
        if(dist < 0.01) {
            gl_FragColor.xyz = vec3(1.0, 0.0, 1.0);
        }
        
    } else {
        gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);
    }
}