#line 1
precision highp float;

#include "shaders/library/common.shader"
#include "shaders/library/random.shader"
#include "shaders/library/brdf.shader"
#include "shaders/library/layer.shader"
#include "shaders/library/covariance.shader"
#line 8

// Uniform variables
uniform vec3  wi;

uniform float  u_Scale;
uniform float  u_Gamma;

uniform float  u_Eta0;
uniform float  u_Eta1;
uniform float  u_Eta2;

uniform float  u_KappaR;
uniform float  u_KappaG;
uniform float  u_KappaB;

// Pass number for the progressive renderer
uniform int u_PassNumber;
uniform int u_ComputationMethod;

#define NB_SAMPLES 1024
#define NB_PER_RUN 1024

/* Gamma transformation
*/
vec3 Gamma(vec3 R) {
   return exp(u_Gamma * log(u_Scale * R));
}

void main(void) {

   gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);

   // Get the omega_o direction from the vPos variable
   vec3 wo;
   wo.x = sin(0.25*PI*(1.0+vPos.x));
   wo.z = sqrt(1.0 - sqr(wo.x));
   wo.y = 0.0;

   // Get the roughness stored in vPos.y
   float alpha = max(0.5*(vPos.y + 1.0), 0.01);

   /* Reference computation */
   if(u_ComputationMethod == 0) {
      for(int i=0; i<NB_PER_RUN; ++i) {
         vec2 rnd; vec3 m1, m2; float pdf;

         /* Radom number */
         int  idx = mod(u_PassNumber*NB_PER_RUN+i, NB_SAMPLES);
         rnd = QMC_Davenport(idx, NB_SAMPLES);

         /* Generate a sample from the bottom GGX distribution
         */
         GGX_Sample_VNDF(wo, max(0.5*(vPos.y + 1.0), 0.01), rnd.xy, m1, pdf);
         vec3 wi = -reflect(wo, m1);
         if(wi.z <= 0.0) { continue; }

         /* Evaluate the Fresnel term for the base layer */
         vec3 R;
         R.x = FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaR);
         R.y = FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaG);
         R.z = FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaB);

         /* Evaluate the Fresnel term for the clear coat */
         R *= 1.0 - FresnelUnpolarized(wi.z, u_Eta1, u_Eta0);

         /* Accumulate in FragColor */
         gl_FragColor.xyz += R * GGX_G1(wi.z, alpha);
      }

   /* Decorrelated integrals */
   } else if(u_ComputationMethod == 1) {
      vec3  R = vec3(0,0,0);
      float T = 0.0;
      float e = 0.0;

      for(int i=0; i<NB_PER_RUN; ++i) {
         /* Random number */
         int  idx = mod(u_PassNumber*NB_PER_RUN+i, NB_SAMPLES);
         vec2 rnd = QMC_Davenport(idx, NB_SAMPLES);


         /* Evaluate the mean Fresnel for the bottom layer */
         vec3 m1, m2; float pdf;
         GGX_Sample_VNDF(wo, max(0.5*(vPos.y + 1.0), 0.01), rnd.xy, m1, pdf);
         vec3 ws = -reflect(wo, m1);
         if(ws.z <= 0.0) { continue; }

         // BOTTOM layer Fresnel evaluation
         float G = GGX_G1(ws.z, alpha);
         R.x += FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaR) * G;
         R.y += FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaG) * G;
         R.z += FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaB) * G;


         /* Evaluate the Fresnel for the TOP layer */
         T += (1.0 - FresnelUnpolarized(ws.z, u_Eta1, u_Eta0)) * G;
         e += G;
      }
      gl_FragColor.xyz = R * (T/e);

   /* Point-wise evaluation of the top interface transmittance using the mean
    * direction wo.
    */
   } else {

      for(int i=0; i<NB_PER_RUN; ++i) {
         /* Radom number
         */
         int  idx = mod(u_PassNumber*NB_PER_RUN+i, NB_SAMPLES);
         vec2 rnd; vec3 m1, m2; float pdf;

         /* Generate a sample from the bottom GGX distribution
         */
         rnd = QMC_Davenport(idx, NB_SAMPLES);
         GGX_Sample_VNDF(wo, max(0.5*(vPos.y + 1.0), 0.01), rnd.xy, m1, pdf);
         vec3 ws = -reflect(wo, m1);
         if(ws.z <= 0.0) { continue; }

         /* Evaluate transmitted Fresnel term, assuming a clear coat above */
         vec3 value = vec3(0,0,0);
         if(ws.z > 0.0) {
            /* Evaluate the Fresnel term for the base layer */
            vec3 R;
            R.x = FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaR);
            R.y = FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaG);
            R.z = FresnelUnpolarized(dot(wo, m1), u_Eta1, u_Eta2, u_KappaB);

            /* Accumulate in FragColor */
            gl_FragColor.xyz += R * GGX_G1(ws.z, alpha);
         }
      }
      gl_FragColor.xyz *= (1.0 - FresnelUnpolarized(wo.z, u_Eta1, u_Eta0));
   }

   gl_FragColor.xyz = Gamma(gl_FragColor.xyz / float(NB_PER_RUN));
   gl_FragColor.w = 1.0;
}
